from bamboo_lib.connectors.models import Connector


def find_param(pipeline_params, key):
    candidates = [x for x in pipeline_params if x.name == key]
    return candidates[0] if candidates else None


def extract(pipeline_params, raw_params):
    '''Given raw form data, coerece parameters to their appropriate types'''
    new_result = {}
    dtype_lookup = {p.name: p.dtype for p in pipeline_params}
    for key, value in raw_params:
        raw_value = dtype_lookup[key]
        # raise Exception(pipeline_params, key )
        param_obj = find_param(pipeline_params, key)
        if param_obj.dtype is Connector:
            value = Connector.fetch(None, value)
        else:
            value = raw_value(value)
        if key not in new_result:
            res = value
        elif isinstance(raw_params[key], list):
            res = raw_params[key] + [value]
        else:
            res = [raw_params[key]] + [value]
        new_result[key] = res
    return new_result



import os
import random
import string
from bamboo_lib.connectors.models import Connector


def grab_connector(file_starting_point, cname):
    """ Given a filepath and connector name, retrieves and builds the
    connector configuration object.

    :param file_starting_point: filepath from which to base the search for the conns.yaml file.
    :param cname: name of the connector to look up
    :return: Connector object
    """
    # try local file first then fall back to global
    par_dir = os.path.abspath(os.path.join(file_starting_point, os.pardir))
    local_conn_path = os.path.join(par_dir, "conns.yaml")
    # source = local_conn_path.get(cname, global_conn_path.get(cname))
    try:
        connector = Connector.fetch(cname, open(local_conn_path))
    except ValueError:
        BAMBOO_FALLBACK_CONNS = os.environ.get("BAMBOO_FALLBACK_CONNS")
        # TODO allow env var to customize default fallback config path
        global_conn_path = os.path.expanduser(os.path.expandvars(BAMBOO_FALLBACK_CONNS))
        connector = Connector.fetch(cname, open(global_conn_path))
    return connector


def random_char(num_characters):
    """ Returns a string of num_characters random ASCII characters.

    :param num_characters: Integer representing the number of characters to appear in the output.
    :return: String
    """
    return ''.join(random.choice(string.ascii_letters) for x in range(num_characters))



import sqlalchemy
from bamboo_lib.connectors.models import BaseDriver


class DatabaseDriver(BaseDriver):
    TYPE = 'SQL Alchemy Driver'

    def __init__(self, **kwargs):
        super(DatabaseDriver, self).__init__(**kwargs)
        self.engine = sqlalchemy.create_engine(self.uri)

    def write_df(self, table_name, df, **kwargs):
        return df.to_sql(table_name,
                         self.engine,
                         **kwargs)



import pandas as pd
import os
from bamboo_lib.models import PipelineStep, ComplexPipelineExecutor  # ,Connector
from bamboo_lib.steps import LoadStep, UnzipStep
from bamboo_lib.connectors.models import Connector
from bamboo_lib.logger import logger
from bamboo_lib.pipelines.pums.mondrian import gen_schema
from bamboo_lib.pipelines.pums.cleaning import cleaner
from bamboo_lib.pipelines.pums.cleaning.util import num_helper
from pandas.api.types import is_numeric_dtype
import numpy as np

MEASURES = ["WKW", "AGEP", "WAGP", "WKHP", "PERNP", "PINCP"]

global file_counter

file_counter = 0


class SaveStep(PipelineStep):
    def __init__(self, **kwargs):
        super(SaveStep, self).__init__(**kwargs)
        self.file_counter = 0

    def run_step(self, prev, params):
        logger.debug("Generating schema...")
        file_path = "$HOME/pums-data/pums_{}{}_{}_{}.csv".format(params.get("regional-mode", "us"), params.get("year"), params.get("estimate"), self.file_counter)
        file_path = os.path.expandvars(file_path)
        prev.to_csv(file_path, index=False)
        self.file_counter += 1
        return file_path


class MakeSchemaStep(PipelineStep):
    def __init__(self, table_name, **kwargs):
        super(MakeSchemaStep, self).__init__(**kwargs)
        self.table_name = table_name

    def run_step(self, prev, params):
        logger.debug("Generating schema...")
        return gen_schema(self.table_name, output_as_fo=True)


class TransferToServerStep(PipelineStep):
    def run_step(self, prev, params):
        logger.debug("Transfering schema...")
        estimate = params["estimate"]
        fpath = "datausa-acs-temp/mondrian/frags/pums_{}.xml".format(estimate)
        self.connector.send_file(prev, fpath, use_fo=True)
        logger.debug("Flushing Mondrian...")
        http_res = self.http_connector.hit()
        logger.debug("Response {} {}".format(http_res.status_code, http_res.json()))
        return True


class ExtractStep(PipelineStep):
    def run_step(self, prev, params):
        logger.debug("Running ExtractStep step...")
        downloaded_file = self.connector.download(params=params)
        return downloaded_file


class AddPkStep(PipelineStep):
    def __init__(self, table_name, **kwargs):
        super(AddPkStep, self).__init__(**kwargs)
        self.table_name = table_name

    def run_step(self, prev, params):
        pk_cols = self.pk_cols
        return self.connector.add_pk(self.schema, self.table_name, pk_cols)


class ChunkStep(PipelineStep):
    def run_step(self, prev, params):
        logger.debug("Running TransformStep step...")
        used_vars = [
            "PUMA", "PWGTP", "SEX", "ESR", "ADJINC",
            "POBP", "SCHL", "ST", "RAC1P", "FOD1P",
            "SERIALNO", "SPORDER", "NAICSP", "SOCP",
            "CIT", "NATIVITY", "HISP", "VPS"
        ]

        weights = ["pwgtp{}".format(i) for i in range(1, 81)]

        if params["estimate"] == 5 or params["year"] == 2017:
            weights = [w.upper() for w in weights]

        cols = used_vars + weights + MEASURES

        def valid_col(x):
            if x in ["PUMA00", "PUMA10", "SOCP00", "SOCP10", "SOCP12", "SERIALNO", "serialno", "POBP05", "POBP12", "pobp05", "pobp12"]:
                return True
            return (x in cols) or (x in [c.lower() for c in cols])
        return pd.read_csv(prev, chunksize=50000000, usecols=valid_col, converters={
            "SERIALNO": str,
            "ESR": str,
            "SEX": str,
            "CIT": str,
            "HISP": str,
            "VPS": str,
            "PUMA": str,
            "PUMA00": str,
            "PUMA10": str,
            "SCHL": str,
            "NATIVITY": str,
            "ST": str,
            "FOD1P": str,
            "NAICSP": str,
            "SOCP": str,
            "SOCP00": str,
            "SOCP10": str,
            "MSP": str,
            "SPORDER": str})
            # na_values=['', ' '])


class TransformStep(PipelineStep):
    def __init__(self, connector, **kwargs):
        super(TransformStep, self).__init__(**kwargs)
        self.connector = connector

    def run_step(self, prev, params):
        df = prev
        estimate = params['estimate']

        if "year" in params:
            df["year"] = params["year"]
        inflation_cols = ["INTP", "OIP", "PAP", "PERNP", "PINCP", "RETP", "SEMP", "SSIP", "SSP", "WAGP"]
        df.ADJINC = df.ADJINC / 1000000.0
        for col in inflation_cols:
            if col in df.columns:
                if not is_numeric_dtype(df[col]):
                    df = num_helper(df, estimate, col)
                df[col] = df[col] * df.ADJINC
        del df['ADJINC']

        df.columns = [col.lower() for col in df.columns]
        # for i in range(1, 81):
            # del df["pwgtp{}".format(i)]
        df.replace('', np.nan, inplace=True)
        if params['estimate'] == 5:
            df = cleaner.standardize(df, **params)
        else:
            df['puma'] = "79500US" + df['st'].astype(str).str.zfill(2) + df['puma'].astype(str).str.zfill(5)
            df['st'] = "04000US" + df['st'].astype(str).str.zfill(2)

        cutoffs = [0, 10000, 20000, 30000, 40000, 50000,
                   60000, 70000, 80000, 90000, 100000, 110000, 120000,
                   130000, 140000, 150000, 160000, 170000, 180000, 190000,
                   200000, 1000000000]
        # TODO testing
        df['wage_bin'] = pd.cut(df.wagp, cutoffs, labels=[
            '< $10K',
            '$10-20k',
            '$20-30k',
            '$30-40k',
            '$40-50k',
            '$50-60k',
            '$60-70k',
            '$70-80k',
            '$80-90k',
            '$90-100k',
            '$100-110k',
            '$110-120k',
            '$120-130k',
            '$130-140k',
            '$140-150k',
            '$150-160k',
            '$160-170k',
            '$170-180k',
            '$180-190k',
            '$190-200k',
            '$200k+'], include_lowest=True)
        err_df_chk = df[(df.wagp >= 200000) & (df.wage_bin.isnull())]
        # TODO some values >200000 showed up as NA. todo, test values >= 200k
        assert err_df_chk.empty, "Bad wage bins!"
        df.wage_bin = df.wage_bin.astype(str)
        df['in_workforce'] = (df.agep.astype(int) >= 16) & (df.wagp > 0) & df.esr.astype(str).isin(["1", "2", "4", "5"])
        # TODO review and test
        df.wkhp = df.wkhp.astype(str).str.strip()
        df.loc[df.wkhp == '', 'wkhp'] = None
        df.wkhp = df.wkhp.astype(float)
        df.loc[df.in_workforce & (df.wkhp >= 35), 'time_status'] = 1
        df.loc[df.in_workforce & (df.wkhp < 35), 'time_status'] = 2
        df['fiveplus'] = df.agep >= 5 ## TODO double check for 2018
        df['birthplace_in_us'] = df.pobp.astype(int) <= 72 ## TODO: Test this!

        df.wage_bin = df.wage_bin.astype(str)
        df.loc[df.wagp.isnull(), 'wage_bin'] = 'N/A'
        wage_bin_map = {
            'N/A': None,
            '< $10K': 1,
            '$10-20k': 2,
            '$20-30k': 3,
            '$30-40k': 4,
            '$40-50k': 5,
            '$50-60k': 6,
            '$60-70k': 7,
            '$70-80k': 8,
            '$80-90k': 9,
            '$90-100k': 10,
            '$100-110k': 11,
            '$110-120k': 12,
            '$120-130k': 13,
            '$130-140k': 14,
            '$140-150k': 15,
            '$150-160k': 16,
            '$160-170k': 17,
            '$170-180k': 18,
            '$180-190k': 19,
            '$190-200k': 20,
            '$200k+': 21
        }
        # raise Exception(df.wage_bin.dtype)
        df['wage_bin_id'] = df.wage_bin.map(wage_bin_map)
        if 'socp00' in df.columns:
            del df['socp00']
        regional_mode = params.get("regional-mode", "us")
        df['nation_id'] = '01000US' if regional_mode == 'us' else  '01000PR'
        # SERIALNO overflows integer as of 2017. Need to fix, but for now since it is not used. will set to 0
        df['serialno'] = 0
        return df


def run(params, **kwargs):
    print(params, "PARAMS!!!!!!!")
    BAMBOO_LIB_DIR = os.environ.get("BAMBOO_LIB_DIR")
    conn_path = os.path.join(BAMBOO_LIB_DIR, "example", "conns.yaml")
    # TODO allow client to pick connectors from UI
    # fallback_source = open(conn_path)
    # connector = Connector.fetch("voyaguers-remote", params.get("output-db", fallback_source))
    # connector = Connector.fetch("postgres-local", open(conn_path))
    regional_mode = params.get("regional-mode", "us")
    if regional_mode == "pr":
        connector2 = Connector.fetch("pums-pr", open(conn_path))
    else:
        connector2 = Connector.fetch("pums-local", open(conn_path))
    # du_http_connector = Connector.fetch("datausa-cube-remote", open(conn_path))
    # dt = {"serialno": "text",
    #       "year": "smallint",
    #       "esr": "smallint",
    #       "cit": "smallint",
    #       "hisp": "smallint",
    #       "nativity": "smallint",
    #       "sex": "smallint",
    #       "puma": "VARCHAR(14)",
    #       "agep": "smallint",
    #       "socp": "VARCHAR(6)",
    #       "naicsp": "VARCHAR(8)",
    #       "pobp": "smallint",
    #       "rac1p": "smallint",
    #       "vps": "smallint",
    #       "msp": "smallint",
    #       "schl": "smallint",
    #       "sporder": "smallint",
    #       "fod1p": "VARCHAR(8)",
    #       "socp": "VARCHAR(6)",
    #       "st": "VARCHAR(9)",
    #       "wage_bin_id": "smallint"}
    dt = {
        "serialno": "text",
        "sporder": "integer",
        "st": "varchar(9)",
        "pwgtp": "integer",
        "agep": "smallint",
        "cit": "smallint",
        "schl": "smallint",
        "sex": "integer",
        "wagp": "real",
        "wkhp": "smallint",
        "wkw": "smallint",
        "esr": "smallint",
        "fod1p": "smallint",
        "hisp": "smallint",
        "naicsp": "varchar(8)",
        "nativity": "smallint",
        "pernp": "real",
        "pincp": "real",
        "rac1p": "smallint",
        "socp10": "varchar(6)",
        "socp": "varchar(6)",
        "vps": "smallint",
        "year": "smallint",
        "puma": "varchar(14)",
        "pobp": "smallint",
        "wage_bin": "varchar(14)",
        "in_workforce": "boolean",
        "time_status": "smallint",
        "fiveplus": "boolean",
        "wage_bin_id": "smallint",
        "nation_id": "varchar(7)",
        "birthplace_in_us": "boolean"
    }

    weights = ["pwgtp{}".format(i) for i in range(1, 81)]
    estimate = params['estimate']
    for w in weights:
        dt[w] = "smallint"
    step1 = ExtractStep(connector=connector2)
    step2 = UnzipStep(pattern=r"\.csv$")
    table_name = "pums_{}_v2".format(estimate)
    chunkstep = ChunkStep()
    step3 = TransformStep(None)
    step4 = SaveStep()
    # step4 = LoadStep(table_name, connector, if_exists="append",
                     # index=False, dtype=dt, schema="pums")
    logger.info("* PUMS pipeline starting...")
    # add_pk_step = AddPkStep(table_name, schema="pums", connector=connector, pk_cols=["sporder", "serialno"])
    pp = ComplexPipelineExecutor(params)
    only_gen_schema = params.get('only-gen-schema', False)
    # schema_step1 = MakeSchemaStep(table_name)
    # schema_connector = Connector.fetch("schema-remote", open(conn_path))
    # schema_step2 = TransferToServerStep(connector=schema_connector, http_connector=du_http_connector)
    if not only_gen_schema:
        pp = pp.next(step1).foreach(step2).foreach(chunkstep).next(step3).next(step4).endeach().endeach()
    # pp = pp.next(schema_step1)
    # pp = pp.next(schema_step2)
    # if not only_gen_schema:
    # #### pp = pp.next(add_pk_step)
    pp.run_pipeline()


if __name__ == '__main__':
    run({"year": 2017, "estimate": 5, "only-gen-schema": False, "regional-mode": "us"})



from bamboo_lib.logger import logger
from bamboo_lib.pipelines.pums.cleaning.puma_helper import update_puma

PUMA = 'puma'
PUMA00 = 'puma00'
PUMA10 = 'puma10'


def _make_geo_id(df, on_col):
    df[on_col] = df[on_col].astype(int)
    mycond = df[on_col] != -9
    df.loc[mycond, on_col] = df.loc[mycond, 'st'].astype(str).str.zfill(2) + df.loc[mycond, on_col].astype(str).str.zfill(5)
    df.loc[~mycond, on_col] = None
    return df


def puma_cleaning(df, year, estimate):
    logger.info("Cleaning PUMAs...")

    if "puma00" in df.columns and "puma10" in df.columns:
        df = _make_geo_id(df, PUMA00)
        df = _make_geo_id(df, PUMA10)
        # apply crosswalk and normalize column name
        df = update_puma(df, PUMA00, True)
        df[PUMA] = '79500US' + df[PUMA]
    else:
        df[PUMA] = "79500US" + df.st + df[PUMA]

    return df



from data_catapult.mondrian.schema import Cube, Schema, Dimension, Measure, Level
from data_catapult.mondrian.schema import CalculatedMember  # DimensionUsage
from io import StringIO


def gen_schema(table_name, output_as_fo=False):
    schema = Schema("datausa")

    cube = Cube(
        name=table_name,
        table=table_name,
        schema="pums",
    )

    cube.add_measure(Measure(
        name="Population Total",
        column="pwgtp",
        aggregator="sum",
        database="postgres",
    ))

    cube.add_measure(Measure(
        name="Age",
        column="agep",
        aggregator="median",
        database="postgres",
    ))

    cube.add_measure(CalculatedMember(
        name="Weighted Wage",
        dimension="Measures",
        formula="[Measures].[Wage]",
        visible=True,
    ))

    dimension = Dimension("Geography")
    dimension.add_level(Level(
        name="State",
        column="st",
        unique_members=True,
    ))
    dimension.add_level(Level(
        name="PUMA",
        column="puma",
        unique_members=True,
    ))
    cube.add_dimension(dimension)

    dimension = Dimension("Year")
    dimension.add_level(Level(
        name="Year",
        column="year",
        unique_members=True,
    ))
    cube.add_dimension(dimension)

    dimension = Dimension("Occupation")
    dimension.add_level(Level(
        name="Occupation",
        column="socp",
        unique_members=True,
    ))
    cube.add_dimension(dimension)

    dimension = Dimension("Workforce Status")
    dimension.add_level(Level(
        name="Workforce Status",
        column="esr",
        unique_members=True,
    ))
    cube.add_dimension(dimension)

    dimension = Dimension("Industry")
    dimension.add_level(Level(
        name="Industry",
        column="naicsp",
        unique_members=True,
    ))
    cube.add_dimension(dimension)

    dimension = Dimension("Race")
    dimension.add_level(Level(
        name="Race",
        column="rac1p",
        unique_members=True,
    ))
    cube.add_dimension(dimension)

    dimension = Dimension("Field of Degree")
    dimension.add_level(Level(
        name="Field of Degree",
        column="fod1p",
        unique_members=True,
    ))
    cube.add_dimension(dimension)

    dimension = Dimension("Place of Birth")
    dimension.add_level(Level(
        name="Place of Birth",
        column="pobp",
        unique_members=True,
    ))
    cube.add_dimension(dimension)

    dimension = Dimension("Sex")
    dimension.add_level(Level(
        name="Sex",
        column="sex",
        unique_members=True,
    ))

    dimension = Dimension("Education")
    dimension.add_level(Level(
        name="Education Level",
        column="schl",
        unique_members=True,
    ))

    cube.add_dimension(dimension)
    schema.add_cube(cube)

    xml_output = schema.xml_render(pretty_print=True)
    if output_as_fo:
        return StringIO(xml_output)
    return xml_output


if __name__ == '__main__':
    print(gen_schema(1))



import paramiko
import os

from bamboo_lib.connectors.models import BaseDriver
from bamboo_lib.logger import logger


class FileTransferDriver(BaseDriver):
    TYPE = 'SCP Secure File Copy Driver'

    def __init__(self, **kwargs):
        self.key_filename = None
        self.port = 22
        self.password = None
        super(FileTransferDriver, self).__init__(**kwargs)
        if self.key_filename:
            self.key_filename = self.expand(self.key_filename)

    @staticmethod
    def expand(fpath):
        return os.path.expanduser(os.path.expandvars(fpath))

    def create_connection(self):
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(self.server, self.port, self.user, self.password, key_filename=self.key_filename)
        return client

    def send_file(self, local_file, remote_file, params=None, use_fo=False):
        logger.info("Starting copy...")
        ssh_client = self.create_connection()
        sftp_client = ssh_client.open_sftp()
        if use_fo and isinstance(local_file, str):
            logger.warn("!WARNING! Trying to send remote file object but string provided")
        if use_fo:
            sftp_client.putfo(local_file, remote_file, confirm=False)
        else:
            local_file = self.expand(local_file)
            sftp_client.put(local_file, remote_file, confirm=False)
        sftp_client.close()
        logger.info("File transfer complete!")
        return True

    def run_command(self, command, params=None):
        logger.info("Starting copy...")
        ssh_client = self.create_connection()
        (stdin, stdout, stderr) = ssh_client.exec_command(command)
        if stderr:
            logger.error("SSH command stderror: \t{}".format(str(stderr)))
        output = [line.strip() for line in stdout.readlines()]
        return output

    def retrieve_file(self, remote_file, local_file, params=None):
        logger.info("Starting copy...")
        local_file = self.expand(local_file)
        ssh_client = self.create_connection()
        sftp_client = ssh_client.open_sftp()
        sftp_client.get(remote_file, local_file)
        ssh_client.close()
        logger.info("File transfer complete!")
        return local_file


if __name__ == '__main__':
    sd = FileTransferDriver(server="35.227.110.181", user='deploy', key_filename="/Users/jspeiser/.ssh/id_rsa_cdc_deploy")
    sd.send_file("/tmp/test.txt", "this_is_now_on_the_server.txt")



from bamboo_lib.connectors.models import BaseDriver
from data_catapult.database.clickhouse import ClickhouseDriver as CatapultCd
from urllib.parse import urlparse


class ClickhouseDriver(BaseDriver):
    """
    Driver that allows bulk data exports to Postgres databases.

    Example usage:

    >>> import pandas as pd
    >>> df = pd.DataFrame({"id": ['a', 'b'], "val": [1, 2]})
    >>> chd = ClickhouseDriver(uri="clickhouse://user:pw@localhost/db_name")
    >>> chd_v2 = ClickhouseDriver(host="localhost", database="db_name", user="user", password="pw")
    >>> chd.write_df("my_new_table", df, pk=["id"])
    """
    TYPE = 'Clickhouse Bulk Write Driver'

    @staticmethod
    def parse_uri(uri):
        """ Checks a URI string that a user passes in and parses it into its components

        :param uri: URI string in format of clickhouse://user:pw@localhost/db_name
        :return: Dictionary with objects for connection settings
        """
        if not uri:
            return None
        result = urlparse(uri)
        host = result.hostname
        username = result.username if result.username else "default"
        password = result.password if result.password else ""

        port = result.port if result.port else 9000
        # use result.path[1:] to skip leading /
        database = result.path[1:] if result.path else "default"
        my_settings = {
            "host": host,
            "port": port,
            "database": database,
            "user": username,
            "password": password
        }
        return my_settings

    def __init__(self, **kwargs):
        super(ClickhouseDriver, self).__init__(**kwargs)
        uri_settings = ClickhouseDriver.parse_uri(kwargs.get("uri", None))
        if not uri_settings:
            self.settings = {
                "host": kwargs.get("host", "localhost"),
                "port": int(kwargs.get("port", 9000)),
                "database": kwargs.get("database", "default"),
                "user": kwargs.get("user", "default"),
                "password": kwargs.get("password", "")
            }
        else:
            self.settings = uri_settings

    def write_df(self, table_name, df, **kwargs):
        chd = CatapultCd(**self.settings)
        # TODO support for schema names and explicit dtypes
        # dtype = kwargs.get("dtype", None)
        return chd.to_sql(df, table_name, **kwargs)



from gen_rca import gen_rca

global design_factor
design_factor = 1.5


def population_measures():
    pop_template = '''
        <Measure name="Total Population{0}" column="pwgtp{0}" aggregator="sum" visible="{1}"/>
    \n'''
    calc_tmp = ""
    for i in range(0, 81):
        val = "" if i == 0 else i
        calc_tmp += pop_template.format(val, "true" if i == 0 else "false")
    return calc_tmp


def population_moe():
    deltas = []
    for i in range(1, 81):
        f = 'Power([Measures].[Total Population] - [Measures].[Total Population{i}], 2)'.format(i=i)
        deltas.append(f)

    summation = "+".join(deltas)
    final_form = "1.645 * Power(.05 * ({}), 0.5)".format(summation)
    return """
    <CalculatedMember name="Total Population MOE" dimension="Measures" visible="false">
        <Annotations>
            <Annotation name="aggregation_method">MOE</Annotation>
            <Annotation name="error_for_measure">Total Population</Annotation>
        </Annotations>
    <Formula>
      <![CDATA[
        {formula}
      ]]>
    </Formula>
    </CalculatedMember>
    <CalculatedMember dimension="Measures" name="Geography Population" visible="false"> <!-- all products, current geo, current year -->
      <Formula>
        <![CDATA[(Measures.[Total Population], [Year].[Year].CurrentMember,
                 [Gender].[Gender].[All Genders],
                 [Age].[Age].[All Ages],
                 [Birthplace].[Birthplace].[All Birthplaces],
                 [Race].[Race].[All Races],
                 [Nativity].[Nativity].[All Nativitys],
                 [Veteran Status].[Veteran Status].[All Veteran Statuss],
                 [Wage Bin].[Wage Bin].[All Wage Bins],
                 [Weeks Worked].[Weeks Worked].[All Weeks Workeds],
                 [Employment Status].[Employment Status].[All Employment Statuss],
                 [PUMS Degree].[PUMS Degree].[All PUMS Degrees],
                 [PUMS Degree Field].[PUMS Degree Field].[All PUMS Degree Fields],
                 [PUMS Occupation].[PUMS Occupation].[All PUMS Occupations],
                 [PUMS Industry].[PUMS Industry].[All PUMS Industrys],
                 [Geography].[Geography].CurrentMember)]]>
      </Formula>
    </CalculatedMember>
    <CalculatedMember name="Total Population MOE Appx" dimension="Measures">
        <Annotations>
            <Annotation name="units_of_measurement">MOE</Annotation>
            <Annotation name="aggregation_method">MOE</Annotation>
            <Annotation name="error_for_measure">Total Population</Annotation>
        </Annotations>
        <Formula>
          <![CDATA[
            1.645 * {design_factor} * Power(99 * [Measures].[Total Population] *
                (1 -
                ([Measures].[Total Population]
                    / [Measures].[Geography Population])), 0.5)
          ]]>
        </Formula>
    </CalculatedMember>

    """.format(formula=final_form, design_factor=design_factor)


def gen_meas(pretty_name, col_name):
    template1 = '''
    <Measure aggregator="None" dataType="Numeric" name="Weighted {0}" visible="false">
      <MeasureExpression>
          <SQL dialect="generic">SUM(pwgtp * {1})</SQL>
      </MeasureExpression>
    </Measure>
    <Measure aggregator="None" dataType="Numeric" name="Weighted {0} SQ" visible="false">
      <MeasureExpression>
          <SQL dialect="monetdb">SUM(pwgtp * {1} * {1})</SQL>
      </MeasureExpression>
    </Measure>
    '''
    tmp1 = template1.format(pretty_name, col_name)
    template2 = '''
    <Measure aggregator="None" dataType="Numeric" name="Weighted {2}{0}" visible="{1}">
      <MeasureExpression>
          <SQL dialect="generic">SUM(pwgtp{0} * {3})</SQL>
      </MeasureExpression>
    </Measure>
    '''
    mea_tmp = ""
    for i in range(1, 81):
        val = "" if i == 0 else i
        # mea_tmp += template2.format(val, "true" if i == 0 else "false", pretty_name, col_name)
        mea_tmp += template2.format(val, "false", pretty_name, col_name)

    return tmp1 + mea_tmp


def get_units(orig_name):
    if orig_name in ["Wage", "Income"]:
        return "USD"
    elif orig_name in ["Age"]:
        return "Years"
    elif orig_name in ["Usual Hours Worked Per Week"]:
        return "Hours"
    else:
        return orig_name

def gen_calcs(pretty_name, col_name):
    units_of_m = get_units(pretty_name)
    template2 = '''
    <CalculatedMember name="Average {0}" dimension="Measures" visible="true">
      <Annotations>
        <Annotation name="units_of_measurement">{3}</Annotation>
        <Annotation name="aggregation_method">AVERAGE</Annotation>
      </Annotations>
      <Formula>[Measures].[Weighted {0}] / [Measures].[Total Population]</Formula>
    </CalculatedMember>
    <CalculatedMember name="Weighted {0} Squared" dimension="Measures" visible="false">
      <Formula>Power([Measures].[Weighted {0}], 2)</Formula>
    </CalculatedMember>
    <CalculatedMember name="SSquared EST {0}" dimension="Measures" visible="false">
          <Formula>
          ([Measures].[Weighted {0} SQ] - (Power([Measures].[Weighted {0}], 2) /
                                          [Measures].[Total Population])) /
            ([Measures].[Total Population] - 1)
          </Formula>
    </CalculatedMember>
    <CalculatedMember name="Average {0} Appx MOE" dimension="Measures">
      <Annotations>
        <Annotation name="units_of_measurement">MOE</Annotation>
        <Annotation name="aggregation_method">MOE</Annotation>
        <Annotation name="error_type">MOE</Annotation>
        <Annotation name="error_for_measure">Average {0}</Annotation>
      </Annotations>
      <Formula>
          IIF([Measures].[Total Population] > 1 AND [Measures].[SSquared EST {0}] > 0, 1.645 * {2} * Power([Measures].[SSquared EST {0}] * (99 / [Measures].[Total Population]), 0.5), NULL)
      </Formula>
    </CalculatedMember>
    '''

    simple_part1 = template2.format(pretty_name, col_name, design_factor, units_of_m)

    template_rw = '''
    <CalculatedMember name="Average {2}{0}" dimension="Measures" visible="{1}">
        <Annotations>
            <Annotation name="units_of_measurement">{2}</Annotation>
            <Annotation name="aggregation_method">AVERAGE</Annotation>
        </Annotations>
      <Formula>[Measures].[Weighted {2}{0}] / [Measures].[Total Population{0}]</Formula>
    </CalculatedMember>\n
    '''

    calc_tmp = ""

    for i in range(1, 81):
        val = "" if i == 0 else i
        calc_tmp += template_rw.format(val, "true" if i == 0 else "false", pretty_name, col_name)

    deltas = []
    for i in range(1, 81):
        f = 'Power([Measures].[Average {pretty_name}] - [Measures].[Average {pretty_name}{i}], 2)'.format(i=i, pretty_name=pretty_name)
        deltas.append(f)
    summation = "+".join(deltas)
    final_form = "1.645 * Power(.05 * ({}), 0.5)".format(summation)
    wagp_moe = """
    <CalculatedMember name="Average {pretty_name} MOE" dimension="Measures" visible="false">
        <Annotations>
            <Annotation name="aggregation_method">MOE</Annotation>
            <Annotation name="error_for_measure">Average {pretty_name}</Annotation>
        </Annotations>
    <Formula>
      <![CDATA[
        {formula}
      ]]>
    </Formula>
    </CalculatedMember>
    """.format(formula=final_form, pretty_name=pretty_name)
    return simple_part1 + calc_tmp + wagp_moe


def gen_all():
    tmp = population_measures()
    metrics = [("Wage", "wagp")]
    metrics += [("Age", "agep")]
    metrics += [("Income", "pincp")]
    metrics += [("Usual Hours Worked Per Week", "wkhp")]

    for metric, col in metrics:
        tmp += gen_meas(metric, col)

    for metric, col in metrics:
        tmp += gen_calcs(metric, col)
    tmp += population_moe()

    tmp += gen_rca("ygopop", "PUMS Occupation", ["Year", "Geography"], ["Employment Status", "Workforce Status", "Age Five Plus", "Employment Time Status"])
    tmp += gen_rca("ygipop", "PUMS Industry", ["Year", "Geography"], ["Employment Status", "Workforce Status", "Age Five Plus", "Employment Time Status"])
    tmp += gen_rca("yocpop", "PUMS Degree Field", ["Year", "PUMS Occupation"], ["Employment Status", "Workforce Status", "Age Five Plus", "Employment Time Status"])
    tmp += gen_rca("yiopop", "PUMS Occupation", ["Year", "PUMS Industry"], ["Employment Status", "Workforce Status", "Age Five Plus", "Employment Time Status"])
    tmp += gen_rca("ygbpop", "Birthplace", ["Year", "Geography"], ["Employment Status", "Workforce Status", "Age Five Plus", "Employment Time Status"])
    tmp += gen_rca("ygcpop", "PUMS Degree Field", ["Year", "Geography"], ["Employment Status", "Workforce Status", "Age Five Plus", "Employment Time Status"])
    tmp += gen_rca("ycbpop", "Birthplace", ["Year", "PUMS Degree Field"], ["Employment Status", "Workforce Status", "Age Five Plus", "Employment Time Status", "Nativity"])

    to_hide = ["ygbpop RCA", "ygopop RCA", "ygipop RCA", "yocpop RCA", "yiopop RCA", "ycbpop RCA"]
    return tmp, to_hide



'''
RCA is just
(a / b)
   /
(c / d)

'''


def gen_members(prefix, title, fn, members, allowed_cuts=None, metric="Total Population"):
    allowed_cuts = [] if not allowed_cuts else allowed_cuts
    my_list = []
    for m in members:
        if fn(m) or m in allowed_cuts:
            my_list.append("[{0}].[{0}].CurrentMember".format(m))
        else:
            my_list.append("[{0}].[{0}].[All {0}s]".format(m))
    my_member_str = ",\n\t\t".join(my_list)
    return """
    <CalculatedMember dimension="Measures" name="{prefix}_rca_{title}_pop" visible="false">
      <Formula>
        <![CDATA[
            (Measures.[{metric}], {my_member_str})
        ]]>
      </Formula>
    </CalculatedMember>
    """.format(prefix=prefix, my_member_str=my_member_str, title=title, metric=metric)

'''
me_mode is the target member
index_modes are the index columns for the RCA calculations
for example to compute the RCA of occupations by year and geography
me_mode = "PUMS Occupation"
and index_modes=["Year", "Geography"]

this function will create variables b,c,d from the RCA formula according the
me/index modes
'''
def simple_rca(prefix, me_mode, index_modes, allowed_cuts=None, metric="Total Population"):
    members = [
        "Year",
        "Geography",
        "Employment Status",
        "Workforce Status",
        "PUMS Occupation",
        "Gender",
        "Age",
        "Race",
        "Nativity",
        "Veteran Status",
        "Wage Bin",
        "Weeks Worked",
        "PUMS Degree",
        "PUMS Degree Field",
        "PUMS Industry"
    ]
    members_b = gen_members(prefix, "b", lambda m: m in index_modes, members, allowed_cuts, metric)
    members_c = gen_members(prefix, "c", lambda m: m in me_mode or m in index_modes[:-1], members, allowed_cuts, metric)
    members_d = gen_members(prefix, "d", lambda m: m in index_modes[:-1], members, allowed_cuts, metric)

    return """
    {members_b}
    {members_c}
    {members_d}

    <CalculatedMember dimension="Measures" name="{prefix} RCA" visible="true">
      <Annotations>
        <Annotation name="aggregation_method">RCA</Annotation>
      </Annotations>
      <Formula>
        <![CDATA[
        IIF([Measures].[{prefix}_rca_c_pop] > 0 and [Measures].[{prefix}_rca_d_pop] > 0,
         ([Measures].[{metric}] / [Measures].[{prefix}_rca_b_pop]) /
         ([Measures].[{prefix}_rca_c_pop] / [Measures].[{prefix}_rca_d_pop])
         , NULL)
        ]]>
      </Formula>
    </CalculatedMember>
    """.format(prefix=prefix, members_b=members_b,
               members_c=members_c, members_d=members_d,
               metric=metric)


if __name__ == "__main__":
    print(gen_rca("ygo", "PUMS Occupation", ["Year", "Geography"], "Employment Status"))




import numpy as np


def num_helper_1yr(df, col):
    df[col] = df[col].astype(str).str.lstrip('0')
    df[col] = df[col].astype(str).str.strip(' ')
    df.loc[df[col].astype(str).str.len() == 0, col] = 0
    df[col] = df[col].astype(float)
    return df


def num_helper_5yr(df, col):
    df[col] = df[col].astype(str).str.strip(' ')
    df.loc[df[col].astype(str).str.len() == 0, col] = 0
    df[col] = df[col].astype(float)
    return df


def num_helper(df, estimate, col):
    if estimate == 1:
        return num_helper_1yr(df, col)
    elif estimate == 5:
        return num_helper_5yr(df, col)
    raise ValueError("Invalid estimate", estimate)


def preclean_io_codes(df):
    to_replace = ["NAICSP02", "NAICSP07", "NAICSP12", "SOCP00", "SOCP10", "SOCP12"]
    for col in to_replace:
        col = col.lower()
        if col in df.columns:
            df.loc[df[col].isin(['N.A.////', 'N.A.//', 'N.A.']), col] = np.nan
    return df



import os
import numpy as np
import pandas as pd
from bamboo_lib.logger import logger
from numpy.random import choice

PUMA = 'puma'
PUMA00 = 'puma00'
PUMA10 = 'puma10'
KATRINA_PUMA = '2277777'
UNKNOWN_LIST = ['XXXXX', None, np.nan, -9]
COL_RATE = "Total Conversion Rate"

RATE_CONVERTER = 'pPUMA00_Pop10'

data_dir = os.path.dirname(__file__)


def get_path(target):
    return os.path.join(data_dir, target)


converters = {"State00": str, "State10": str, PUMA00: str, PUMA10: str}
puma_map = pd.read_excel(get_path('data/PUMA2000_PUMA2010_crosswalk.xls'), converters=converters)[["State00", "State10", PUMA00, PUMA10, "pPUMA00_Pop10"]]
puma_map[PUMA00] = puma_map["State00"] + puma_map[PUMA00]
puma_map[PUMA10] = puma_map["State10"] + puma_map[PUMA10]
puma_map = puma_map.drop(["State00", "State10"], axis=1)
puma_map[RATE_CONVERTER] = puma_map[RATE_CONVERTER] / 100.00
puma_map = puma_map.set_index(PUMA00)


def diluter_update_puma(df, on_col, has_puma10=False):
    '''This verison will distribute the person weight'''
    df = pd.merge(puma_map, df, left_on=PUMA00, right_on=on_col, how="right")
    df.drop([PUMA00], axis=1, inplace=True)
    # df.rename(columns={PUMA10: PUMA}, inplace=True)

    # Dilute person weight values by RATE_CONVERTER
    for col in df.columns:
        if "pwgtp" in col.lower():
            df.loc[df['puma10_x'].notnull(), col] = df.loc[df['puma10_x'].notnull(), col] * df[RATE_CONVERTER]

    df.loc[df['puma10_x'].notnull(), PUMA] = df['puma10_x']
    df.loc[df['puma10_y'].notnull(), PUMA] = df['puma10_y']
    df.drop(["puma10_x", "puma10_y", RATE_CONVERTER], axis=1, inplace=True)
    return df


def randomizer(code):
    if code in UNKNOWN_LIST:
        raise Exception("INVALID PUMA! UNKNOWN!")
    if code == KATRINA_PUMA:
        return KATRINA_PUMA

    tmpdf = puma_map.ix[code]

    if isinstance(tmpdf, pd.Series):
        return tmpdf[PUMA10]

    # print code
    # normalize tmpdf[RATE_CONVERTER].values
    tmp_arr = tmpdf[RATE_CONVERTER].values
    tot_sum = sum(tmp_arr)
    p_vals = [1.0 * p / tot_sum for p in tmp_arr]
    selected_idx = choice(range(len(tmpdf)), p=p_vals)
    return tmpdf.iloc[selected_idx][PUMA10]


def update_puma(df, on_col, has_puma10=False):
    to_drop = [on_col]

    df.loc[df[on_col].notnull(), on_col] = df.loc[df[on_col].notnull(), on_col].apply(randomizer)
    df.loc[df[on_col].notnull(), PUMA] = df.loc[df[on_col].notnull(), on_col]
    df.loc[df[PUMA10].notnull(), PUMA] = df.loc[df[PUMA10].notnull(), PUMA10]
    to_drop.append(PUMA10)
    df.drop(to_drop, axis=1, inplace=True)
    return df


if __name__ == '__main__':
    moi = pd.DataFrame({"puma10": [None, "2500506"], "puma00": ["5100600", None], "PWGTP": [100, 200], "PWGTP1": [100, 200], "PWGTP2": [60, 200]})
    print(update_puma(moi, "puma00"))



# test_capitalize.py
import pytest

from bamboo_lib.models import PipelineStep, ComplexPipelineExecutor, AdvancedPipelineExecutor
from bamboo_lib.steps import LoadStep, DownloadStep

def build_pipeline(pipeline_kind=ComplexPipelineExecutor, complexity="advanced"):

    class Nest1Step(PipelineStep):
        def run_step(self, prev_result, params):
            print("-----Nest1Step-----")
            for item in [3, 5]:
                yield item

    class Nest2Step(PipelineStep):
        def run_step(self, prev_result, params):
            print("-----Nest2Step-----")
            for item in [prev_result * 4, prev_result * 6]:
                yield item

    class DisplayStep(PipelineStep):
        def run_step(self, prev_result, params):
            print("RESULT!!", prev_result, "\n")
            return prev_result

    class DummyStepSimple(PipelineStep):
        def run_step(self, prev_result, params):
            print("-----START-----")
            return "Gamma"

    class DummyStep(PipelineStep):
        def run_step(self, prev_result, params):
            print("-----START-----")
            for item in ["Alpha", "Beta", "Gamma"]:
                yield item

    class TransformStep(PipelineStep):
        def run_step(self, prev_result, params):
            print("Check 1:", prev_result)
            return prev_result + "ADDED"

    class GenerateSchemaStep(PipelineStep):
        def run_step(self, prev_result, params):
            # this should return a df reflecting the changes in the previous step,
            # but instead it returns the initial version of the dataframe object without any changes
            title = prev_result
            print("Check 2: (Should be +ADDED)", title)
            print("----END----")
            return title

    transform_step = TransformStep()
    generate_schema_step = GenerateSchemaStep()
    pp = pipeline_kind({})
    if complexity == "advanced":
        dummy_step = DummyStep()
        pp = pp.foreach(dummy_step).next(transform_step).next(generate_schema_step).endeach()
    elif complexity == "nested":
        nest_1_step = Nest1Step()
        nest_2_step = Nest2Step()
        display_step = DisplayStep()
        pp = pp.foreach(nest_1_step).foreach(nest_2_step).next(display_step).endeach().endeach()
    else:
        dummy_step_simple = DummyStepSimple()
        pp = pp.next(dummy_step_simple).next(transform_step).next(generate_schema_step)

    return pp


def test_complex_pipeline():
    pipeline = build_pipeline(ComplexPipelineExecutor)
    res = pipeline.run_pipeline()
    assert res == "GammaADDED"


def test_complex_pipeline_with_nested_loops():
    pipeline = build_pipeline(ComplexPipelineExecutor, complexity="nested")
    assert isinstance(pipeline, ComplexPipelineExecutor)
    res = pipeline.run_pipeline()
    assert res == 30


def test_advanced_pipeline_without_loops():
    pipeline = build_pipeline(AdvancedPipelineExecutor, complexity="simple")
    res = pipeline.run_pipeline()
    assert res == "GammaADDED"


def test_advanced_pipeline_with_loops():
    pipeline = build_pipeline(AdvancedPipelineExecutor, complexity="advanced")
    res = pipeline.run_pipeline()
    assert res == "GammaADDED"


def test_advanced_pipeline_with_nested_loops():
    pipeline = build_pipeline(AdvancedPipelineExecutor, complexity="nested")
    res = pipeline.run_pipeline()
    assert res == 30

def test_create_load_step_with_nullable_list():
    load_step = LoadStep("bls_test_import", None, if_exists="append", pk=["year"], nullable_list=["alpha"])
    assert load_step

def test_callback_in_dl_step():
    def my_callback_fun(x):
        return "This is it."
    dl_step = DownloadStep(connector="http-local", connector_path=__file__, callback=my_callback_fun)
    res = dl_step.connector.download(callback=my_callback_fun, force=True)
    with open(res, 'r') as my_file:
        assert my_file.read() == "This is it."

def test_no_callback_in_dl_step():
    dl_step = DownloadStep(connector="http-local", connector_path=__file__)
    res = dl_step.connector.download(force=True)
    with open(res, 'r') as my_file:
        assert '''"title": "delectus aut autem"''' in my_file.read()



from bamboo_lib.models import Parameter, BasePipeline
from bamboo_lib.pipelines.pums.pums_runner import run
from bamboo_lib.connectors.models import Connector


class PumsPipeline(BasePipeline):
    @staticmethod
    def pipeline_id():
        return 'pums'

    @staticmethod
    def name():
        return 'PUMS Pipeline'

    @staticmethod
    def description():
        return 'Processes data from PUMS'

    @staticmethod
    def website():
        return 'https://www.census.gov/programs-surveys/acs/data/pums.html'

    @staticmethod
    def parameter_list():
        return [
            Parameter(name="year", dtype=int, allow_multiple=False, options=[2013, 2014, 2015, 2016]),
            Parameter(name="estimate", dtype=int, options=[1, 5]),
            Parameter(name="regional-mode", dtype=bool, options=["us", "pr"]),
            Parameter(label="Save results in database", name="write-to-db", dtype=bool, options=[True, False]),
            Parameter(label="Only generate schema", name="only-gen-schema", dtype=bool),
            Parameter(label="Output database connector", name="output-db", dtype=str, source=Connector)
        ]

    @staticmethod
    def run(params_dict, **kwargs):
        return run(params_dict)




def VBase(object):
    def __init__(self):
        self.children = None

    def attach(self, node):
        self.children.append(node)


def Cube(VBase):
    def __init__(self, name, physical_table):
        super(Cube, self).__init__(self)
        self.name = name
        self.physical_table = physical_table


def Dimension(VBase):
    def __init__(self, name, physical_table=None, foreignKey=None):
        self.name = name
        self.physical_table = physical_table
        self.foreignKey = foreignKey


def Hierarchy(VBase):
    def __init__(self, name, hasAll=True):
        self.name = name
        self.hasAll = hasAll


def Level(VBase):
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            if key not in ["name", "nameColumn", "column", "ordinal", "uniqueMembers"]:
                raise ValueError("bad value for key", key)
            setattr(self, key, val)


def Measure(VBase):
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)


def CalculatedMember(VBase):
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)



import os
from sshtunnel import SSHTunnelForwarder
from bamboo_lib.connectors.models import BaseDriver


class SSHTunnelDriver(BaseDriver):
    '''
    Example Usage:
    sd = SSHTunnelDriver(host="canon-api.datausa.io", user="deploy", ssh_pkey="$HOME/.ssh/id_rsa_dw_deploy")
    sd.start()
    '''
    TYPE = 'Redis Connection Driver'

    def __init__(self, **kwargs):
        super(SSHTunnelDriver, self).__init__(**kwargs)
        host = kwargs.get("host", "localhost")
        local_port = kwargs.get("local_port", 6379)
        user = kwargs.get("user", "root")
        pw = kwargs.get("password", None)
        remote_port = kwargs.get("remote_port", 6379)
        ssh_pkey = kwargs.get("ssh_pkey", None)
        if ssh_pkey:
            ssh_pkey = os.path.expanduser(os.path.expandvars(ssh_pkey))
        ssh_private_key_password_env_var = kwargs.get("ssh_private_key_password_env_var", None)
        self.remote_server = SSHTunnelForwarder(
            host,
            ssh_username=user,
            ssh_password=pw,
            remote_bind_address=("127.0.0.1", remote_port),
            local_bind_address=("0.0.0.0", local_port),
            ssh_pkey=ssh_pkey,
            ssh_private_key_password=ssh_private_key_password_env_var
        )

    def start(self):
        return self.remote_server.start()

    def close(self):
        return self.remote_server.close()



from bamboo_lib.pipelines.pums.cleaning.util import num_helper, preclean_io_codes
from bamboo_lib.pipelines.pums.cleaning.puma import puma_cleaning
from bamboo_lib.pipelines.pums.cleaning.io_converter import occ_convert


def clean_pobp(df):
    print("Cleaning birthplaces....")
    if 'pobp05' in df.columns:
        pobp05i = df['pobp05'].astype(int)
        pobp12i = df['pobp12'].astype(int)

        df.loc[pobp05i >= 0, 'pobp'] = df.loc[pobp05i >= 0, 'pobp05']
        df.loc[pobp12i >= 0, 'pobp'] = df.loc[pobp12i >= 0, 'pobp12']
        del df['pobp05']
        del df['pobp12']
    print("Completed cleaning birthplaces!")
    return df


def standardize(df, **kwargs):
    estimate = kwargs.get('estimate')
    year = kwargs.get('year')

    df = num_helper(df, estimate, 'wagp')
    df = num_helper(df, estimate, 'wkhp')

    assert df[df.st.isnull()].empty

    df = preclean_io_codes(df)

    df = puma_cleaning(df, year, estimate)
    df["st"] = "04000US" + df.st
    df = occ_convert(df, kwargs)
    df = clean_pobp(df)
    return df


if __name__ == '__main__':
    print("HELLO!!!")



import pkgutil as __pkgutil

__path__ = __pkgutil.extend_path(__path__, __name__)
for __importer, __modname, __ispkg in __pkgutil.walk_packages(path=__path__, prefix=__name__ + '.'):
    __import__(__modname)



import os
import sys
import requests
import hashlib
import io
import time

from bamboo_lib.logger import logger
from bamboo_lib.models import PipelineStep
from bamboo_lib.connectors.models import BaseDriver
from bamboo_lib.exceptions import DownloadFailedException
class HttpDriver(BaseDriver):
    TYPE = 'HTTP Web Driver'

    def __init__(self, **kwargs):
        self.timeout = kwargs.get("timeout", None)
        self.max_retries = kwargs.get("max_retries", None)
        self._num_retries = 0
        self.backoff_time_wait = int(kwargs.get("backoff_time_wait", 5)) # time in seconds
        # self.backoff_strategy = 'EXPONENTIAL'
        super(HttpDriver, self).__init__(**kwargs)
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:62.0) Gecko/20100101 Firefox/62.0'
        }

    def hit(self):
        logger.debug("HttpDriver hitting {}".format(self.uri))
        return requests.get(self.uri, headers=self.headers)

    def save_to_disk(self, io_handle, content, callback, params):
        logger.info("Saving web asset to disk (using callback={})....".format(callback is not None))
        content = io.BytesIO(content)
        if callback:
            if isinstance(callback, PipelineStep):
                content = callback(content, params)
            else:
                content = callback(content)
        if isinstance(content, str):
            content = content.encode("utf-8")
        elif isinstance(content, io.BytesIO):
            content = content.read()
        elif isinstance(content, io.StringIO):
            content = content.getvalue().encode("utf-8")
        io_handle.write(content)

    def download(self, params=None, http_params=None, force=False, callback=None):
        uri = self.uri
        uri = BaseDriver.resolve_params(uri, params).encode('utf-8')
        download_path = os.environ.get("BAMBOO_DOWNLOAD_FOLDER", "/tmp")
        if not os.path.exists(download_path):
            os.makedirs(download_path)
        local_filename = os.path.join(download_path, hashlib.sha224(uri).hexdigest())
        logger.info("Hitting URL: {} ...".format(uri))
        # if already downloaded, dont redownload unless required
        if not force and os.path.isfile(local_filename):
            logger.info("Reading web asset from local cache...")
            return local_filename

        inmemory_buffer = io.BytesIO()
        try:
            response = requests.get(uri, stream=True, headers=self.headers, timeout=self.timeout)
        except requests.exceptions.Timeout:
            logger.error("Request to {} timed out.".format(uri))
            if self.max_retries and self._num_retries < self.max_retries:
                self._num_retries += 1
                # wait
                logger.info("Going to retry (retry #{}) request, but first going to sleep...".format(self._num_retries))
                sleep_time = pow(self.backoff_time_wait, self._num_retries)
                logger.info("Sleeping for {} seconds before retrying...".format(sleep_time))
                time.sleep(sleep_time)
                return self.download(params, http_params, force, callback)
            else:
                raise DownloadFailedException("Download failed. Request timed out and retry limit ({}) exceeded.".format(self.max_retries))
        with open(local_filename, "wb") as f:
            logger.info("Downloading {}".format(uri))
            total_length = response.headers.get('content-length')
            if total_length is None:  # no content length header
                self.save_to_disk(f, response.content, callback, params)
            else:
                dl = 0
                total_length = int(total_length)
                # in the case of the chunked response
                # write everything to buffer, then apply callback fn
                for data in response.iter_content(chunk_size=4096):
                    dl += len(data)
                    inmemory_buffer.write(data)
                    done = int(50 * dl / total_length)
                    sys.stdout.write("\r[%s%s]" % ('=' * done, ' ' * (50 - done)))
                    sys.stdout.flush()
                self.save_to_disk(f, inmemory_buffer.getvalue(), callback, params)
        inmemory_buffer.close()
        return local_filename



import logging
import logging.config
import os

logger_conf_path = os.environ.get("BAMBOO_LOGGER_CONF", None)

# logging.basicConfig(level=logging.DEBUG)
mode = "Default logger"

if logger_conf_path and os.path.isfile(logger_conf_path):
    logging.config.fileConfig(logger_conf_path)
    mode = "Using logger config from BAMBOO_LOGGER_CONF..."
elif os.path.isfile('logging.conf'):
    logging.config.fileConfig('logging.conf')
    mode = "Using logger config from logging.conf file..."
else:
    logging.basicConfig(level=logging.DEBUG)

logger = logging.getLogger(__name__)

if mode != "Default logger":
    logger.info(mode)



# test_capitalize.py
import pytest
import os
import logging

@pytest.fixture(autouse=True)
def env_setup(monkeypatch):
    my_path = os.path.join(__file__, "../", "example", "logging.conf")
    monkeypatch.setenv('BAMBOO_LOGGER_CONF', str(my_path))
    from importlib import reload


# def test_basic_logger():
#     from bamboo_lib.logger import logger
#     assert logger.getEffectiveLevel() == logging.INFO



PUMS_MASTER_XML = """<?xml version="1.0" ?>
<Schema name="datausa">

<Cube name="pums_1">
    <Annotations>
      <Annotation name="source_name">Census Bureau</Annotation>
      <Annotation name="source_description">The American Community Survey (ACS) Public Use Microdata Sample (PUMS) files are a set of untabulated records about individual people or housing units. The Census Bureau produces the PUMS files so that data users can create custom tables that are not available through pretabulated (or summary) ACS data products.</Annotation>
      <Annotation name="dataset_name">ACS PUMS 1-Year Estimate</Annotation>
      <Annotation name="dataset_link">https://census.gov/programs-surveys/acs/technical-documentation/pums.html</Annotation>
      <Annotation name="subtopic">Demographics</Annotation>
      <Annotation name="table_id">PUMS</Annotation>
      <Annotation name="topic">Diversity</Annotation>
      <Annotation name="hidden_measures">{to_hide}</Annotation>
    </Annotations>
  <Table name="pums_1" schema="pums"/>
  <Dimension foreignKey="puma" name="Geography">
    <Annotations>
      <Annotation name="dim_type">GEOGRAPHY</Annotation>
    </Annotations>
    <Hierarchy hasAll="true" primaryKey="geoid" defaultMember="[Nation].[United States]">
      <Table name="pumas" schema="shapes2017"/>
      <Level column="nation_id" nameColumn="nation_name" name="Nation" uniqueMembers="true" hideMemberIf="IfBlankName"/>
      <Level column="state_id" nameColumn="state_name" name="State" uniqueMembers="true"/>
      <Level column="geoid" nameColumn="name" name="PUMA" uniqueMembers="true"/>
    </Hierarchy>
  </Dimension>
  <Dimension name="Year">
    <Annotations>
      <Annotation name="dim_type">TIME</Annotation>
    </Annotations>
    <Hierarchy hasAll="true">
      <Level column="year" levelType="TimeYears" name="Year" type="Numeric" uniqueMembers="true"/>
    </Hierarchy>
  </Dimension>
  <Dimension name="Workforce Status">
    <Hierarchy hasAll="true">
      <Level column="in_workforce" name="Workforce Status" type="Boolean" uniqueMembers="true"/>
    </Hierarchy>
  </Dimension>
  <Dimension foreignKey="time_status" name="Employment Time Status">
    <Hierarchy hasAll="true">
      <InlineTable alias="emp_time_status">
        <ColumnDefs>
          <ColumnDef name="id" type="Numeric"/>
          <ColumnDef name="description" type="String"/>
        </ColumnDefs>
        <Rows>
          <Row>
            <Value column="id">1</Value>
            <Value column="description">Full-time</Value>
          </Row>
          <Row>
            <Value column="id">2</Value>
            <Value column="description">Part-time</Value>
          </Row>
        </Rows>
      </InlineTable>
      <Level column="id" name="Employment Time Status" nameColumn="description" uniqueMembers="true">
      </Level>
    </Hierarchy>
  </Dimension>
  <Dimension name="Age">
    <Hierarchy hasAll="true">
      <Level column="agep" name="Age" type="Numeric" uniqueMembers="true"/>
    </Hierarchy>
  </Dimension>
  <Dimension foreignKey="wkw" name="Weeks Worked">
    <Hierarchy hasAll="true">
    <InlineTable alias="wks">
      <ColumnDefs>
        <ColumnDef name="id" type="Numeric"/>
        <ColumnDef name="description" type="String"/>
      </ColumnDefs>
      <Rows>
        <Row>
          <Value column="id">1</Value>
          <Value column="description">50-52 Weeks</Value>
        </Row>
        <Row>
          <Value column="id">2</Value>
          <Value column="description">48-49 Weeks</Value>
        </Row>
        <Row>
          <Value column="id">3</Value>
          <Value column="description">40-47 Weeks</Value>
        </Row>
        <Row>
          <Value column="id">4</Value>
          <Value column="description">27-39 Weeks</Value>
        </Row>
        <Row>
          <Value column="id">5</Value>
          <Value column="description">14-26 Weeks</Value>
        </Row>
        <Row>
          <Value column="id">6</Value>
          <Value column="description">Less than 14 Weeks</Value>
        </Row>
      </Rows>
    </InlineTable>
    <Level column="id" name="Weeks Worked" nameColumn="description" uniqueMembers="true">
    </Level>
  </Hierarchy>
  </Dimension>
<DimensionUsage foreignKey="socp" source="PUMS Occupation" name="PUMS Occupation" />
<DimensionUsage foreignKey="naicsp" source="PUMS Industry" name="PUMS Industry" />
<DimensionUsage foreignKey="pobp" source="Birthplace" name="Birthplace" />
<DimensionUsage foreignKey="hisp" source="PUMS Ethnicity" name="PUMS Ethnicity" />

<Dimension foreignKey="schl" name="PUMS Degree">
    <Hierarchy hasAll="true" primaryKey="schl">
        <Table name="schl" schema="pums_dims"/>
        <Level column="schl_group" nameColumn="schl_group" name="Group" uniqueMembers="true"/>
        <Level column="schl" nameColumn="schl_name" name="Degree" uniqueMembers="true"/>
    </Hierarchy>
</Dimension>
  <Dimension foreignKey="fod1p" name="PUMS Degree Field">
    <Hierarchy hasAll="true" primaryKey="fod1p">
        <Table name="fod1p" schema="pums_dims"/>
        <Level column="cip2" nameColumn="cip2_name" name="CIP2" uniqueMembers="true"/>
        <Level column="fod1p" nameColumn="fod1p_name" name="FOD1P" uniqueMembers="true"/>
    </Hierarchy>
</Dimension>
  <Dimension foreignKey="sex" name="Gender">
    <Hierarchy hasAll="true">
      <InlineTable alias="sex">
        <ColumnDefs>
          <ColumnDef name="id" type="Numeric"/>
          <ColumnDef name="description" type="String"/>
        </ColumnDefs>
        <Rows>
          <Row>
            <Value column="id">1</Value>
            <Value column="description">Male</Value>
          </Row>
          <Row>
            <Value column="id">2</Value>
            <Value column="description">Female</Value>
          </Row>
        </Rows>
      </InlineTable>
      <Level column="id" name="Gender" nameColumn="description" uniqueMembers="true">
      </Level>
    </Hierarchy>
  </Dimension>
  <Dimension foreignKey="rac1p" name="Race">
    <Hierarchy hasAll="true">
      <InlineTable alias="race">
        <ColumnDefs>
          <ColumnDef name="id" type="Numeric"/>
          <ColumnDef name="description" type="String"/>
        </ColumnDefs>
        <Rows>
          <Row>
            <Value column="id">1</Value>
            <Value column="description">White</Value>
          </Row>
          <Row>
            <Value column="id">2</Value>
            <Value column="description">Black</Value>
          </Row>
          <Row>
            <Value column="id">3</Value>
            <Value column="description">American Indian</Value>
          </Row>
          <Row>
            <Value column="id">4</Value>
            <Value column="description">Alaska Native</Value>
          </Row>
          <Row>
            <Value column="id">5</Value>
            <Value column="description">American Indian and Alaska Native tribes specified; or American Indian or Alaska Native, not specified and no other races</Value>
          </Row>
          <Row>
            <Value column="id">6</Value>
            <Value column="description">Asian</Value>
          </Row>
          <Row>
            <Value column="id">7</Value>
            <Value column="description">Native Hawaiian and Other Pacific Islander</Value>
          </Row>
          <Row>
            <Value column="id">8</Value>
            <Value column="description">Other</Value>
          </Row>
          <Row>
            <Value column="id">9</Value>
            <Value column="description">Two or More Races</Value>
          </Row>
        </Rows>
      </InlineTable>
      <Level column="id" name="Race" nameColumn="description" uniqueMembers="true">
      </Level>
    </Hierarchy>
  </Dimension>

<Dimension foreignKey="esr" name="Employment Status">
    <Hierarchy hasAll="true">
      <InlineTable alias="esr">
        <ColumnDefs>
          <ColumnDef name="id" type="Numeric"/>
          <ColumnDef name="description" type="String"/>
          <ColumnDef name="parent" type="String"/>
        </ColumnDefs>
        <Rows>
          <Row>
            <Value column="id">1</Value>
            <Value column="description">Civilian employed, at work</Value>
            <Value column="parent">Employed</Value>

          </Row>
          <Row>
            <Value column="id">2</Value>
            <Value column="description">Civilian employed, with a job but not at work</Value>
            <Value column="parent">Employed</Value>
          </Row>
          <Row>
            <Value column="id">3</Value>
            <Value column="description">Unemployed</Value>
            <Value column="parent">Not Employed</Value>
          </Row>
          <Row>
            <Value column="id">4</Value>
            <Value column="description">Armed forces, at work</Value>
            <Value column="parent">Employed</Value>
          </Row>
          <Row>
            <Value column="id">5</Value>
            <Value column="description">Armed forces, with a job but not at work</Value>
            <Value column="parent">Employed</Value>
          </Row>
          <Row>
            <Value column="id">6</Value>
            <Value column="description">Not in labor force</Value>
            <Value column="parent">Not Employed</Value>
          </Row>
        </Rows>
      </InlineTable>
      <Level column="parent" name="Employment Status Parent" uniqueMembers="true"></Level>
      <Level column="id" name="Employment Status" nameColumn="description" uniqueMembers="true">
      </Level>
    </Hierarchy>
  </Dimension>
  <Dimension foreignKey="cit" name="Citizenship Status">
    <Hierarchy hasAll="true">
      <InlineTable alias="cit">
        <ColumnDefs>
          <ColumnDef name="id" type="Numeric"/>
          <ColumnDef name="description" type="String"/>
          <ColumnDef name="parent" type="String"/>
        </ColumnDefs>
        <Rows>
          <Row>
            <Value column="id">1</Value>
            <Value column="description">Born in the U.S.</Value>
            <Value column="parent">U.S. Citizen</Value>

          </Row>
          <Row>
            <Value column="id">2</Value>
            <Value column="description">Born in Puerto Rico, Guam, the U.S. Virgin Islands, or the Northern Marianas</Value>
            <Value column="parent">U.S. Citizen</Value>
          </Row>
          <Row>
            <Value column="id">3</Value>
            <Value column="description">Born abroad of American parent(s)</Value>
            <Value column="parent">U.S. Citizen</Value>
          </Row>
          <Row>
            <Value column="id">4</Value>
            <Value column="description">U.S. citizen by naturalization</Value>
            <Value column="parent">U.S. Citizen</Value>
          </Row>
          <Row>
            <Value column="id">5</Value>
            <Value column="description">Not a citizen of the U.S.</Value>
            <Value column="parent">Not a citizen of the U.S.</Value>
          </Row>
        </Rows>
      </InlineTable>
      <Level column="parent" name="Citizenship Status Parent" uniqueMembers="true"></Level>
      <Level column="id" name="Citizenship Status" nameColumn="description" uniqueMembers="true">
      </Level>
    </Hierarchy>
  </Dimension>
  <Dimension foreignKey="nativity" name="Nativity">
    <Hierarchy hasAll="true">
      <InlineTable alias="nativity">
        <ColumnDefs>
          <ColumnDef name="id" type="Numeric"/>
          <ColumnDef name="description" type="String"/>
        </ColumnDefs>
        <Rows>
          <Row>
            <Value column="id">1</Value>
            <Value column="description">Native</Value>
          </Row>
          <Row>
            <Value column="id">2</Value>
            <Value column="description">Foreign born</Value>
          </Row>
        </Rows>
      </InlineTable>
      <Level column="id" name="Nativity" nameColumn="description" uniqueMembers="true">
      </Level>
    </Hierarchy>
  </Dimension>

  <Dimension foreignKey="vps" name="Veteran Status">
    <Hierarchy hasAll="true">
      <InlineTable alias="veteran_status">
        <ColumnDefs>
          <ColumnDef name="id" type="Numeric"/>
          <ColumnDef name="description" type="String"/>
          <ColumnDef name="parent_description" type="String"/>

        </ColumnDefs>
        <Rows>
          <Row>
            <Value column="id">1</Value>
            <Value column="description">Gulf War: 9/2001 or later</Value>
            <Value column="parent_description">Gulf War</Value>
          </Row>
          <Row>
            <Value column="id">2</Value>
            <Value column="description">Gulf War: 9/2001 or later and Gulf War: 8/1990 - 8/2001</Value>
            <Value column="parent_description">Gulf War</Value>
          </Row>
          <Row>
            <Value column="id">3</Value>
            <Value column="description">Gulf War: 9/2001 or later and Gulf War: 8/1990 - 8/2001 and Vietnam Era</Value>
            <Value column="parent_description">Gulf War</Value>
          </Row>
          <Row>
            <Value column="id">4</Value>
            <Value column="description">Gulf War: 8/1990 - 8/2001</Value>
            <Value column="parent_description">Gulf War</Value>
          </Row>
          <Row>
            <Value column="id">5</Value>
            <Value column="description">Gulf War: 8/1990 - 8/2001 and Vietnam Era</Value>
            <Value column="parent_description">Gulf War</Value>
          </Row>
          <Row>
            <Value column="id">6</Value>
            <Value column="description">Vietnam Era</Value>
            <Value column="parent_description">Vietnam Era</Value>
          </Row>
          <Row>
            <Value column="id">7</Value>
            <Value column="description">Vietnam Era and Korean War</Value>
            <Value column="parent_description">Vietnam Era</Value>
          </Row>
          <Row>
            <Value column="id">8</Value>
            <Value column="description">Vietnam Era, Korean War, and WWII</Value>
            <Value column="parent_description">Vietnam Era War</Value>
          </Row>
          <Row>
            <Value column="id">9</Value>
            <Value column="description">Korean War</Value>
            <Value column="parent_description">Korean War</Value>
          </Row>
          <Row>
            <Value column="id">10</Value>
            <Value column="description">Korean War and WWII</Value>
            <Value column="parent_description">Korean War</Value>
          </Row>
          <Row>
            <Value column="id">11</Value>
            <Value column="description">WWII</Value>
            <Value column="parent_description">WWII</Value>
          </Row>
          <Row>
            <Value column="id">12</Value>
            <Value column="description">Between Gulf War and Vietnam Era only</Value>
            <Value column="parent_description">Between Gulf War and Vietnam Era only</Value>
          </Row>
          <Row>
            <Value column="id">13</Value>
            <Value column="description">Between Vietnam Era and Korean War only</Value>
            <Value column="parent_description">Between Vietnam Era and Korean War only</Value>
          </Row>
          <Row>
            <Value column="id">14</Value>
            <Value column="description">Between Korean War and World War II only</Value>
            <Value column="parent_description">Between Korean War and World War II only</Value>
          </Row>
          <Row>
            <Value column="id">15</Value>
            <Value column="description">Pre-WWII only</Value>
            <Value column="parent_description">Pre-WWII only</Value>
          </Row>
        </Rows>
      </InlineTable>
      <Level column="id" name="Veteran Status" nameColumn="description" uniqueMembers="true">
      </Level>
    </Hierarchy>
  </Dimension>

  <Dimension foreignKey="wage_bin" name="Wage Bin">
    <Hierarchy hasAll="true">
      <Level column="wage_bin_id" nameColumn="wage_bin" name="Wage Bin" uniqueMembers="true" ordinalColumn="wage_bin_id"></Level>
    </Hierarchy>
  </Dimension>

  <Dimension name="Age Five Plus">
    <Hierarchy hasAll="true">
      <Level column="fiveplus" name="Age Five Plus" type="Boolean" uniqueMembers="true"/>
    </Hierarchy>
  </Dimension>

  <Measure aggregator="count" column="pwgtp" name="Record Count" visible="true">
    <Annotations>
        <Annotation name="units_of_measurement">Sample Records</Annotation>
    </Annotations>
  </Measure>

  {calculations}
</Cube>
</Schema>"""



import redis

from bamboo_lib.connectors.models import BaseDriver


class RedisDriver(BaseDriver):
    TYPE = 'Redis Connection Driver'

    def __init__(self, **kwargs):
        super(RedisDriver, self).__init__(**kwargs)
        host = kwargs.get("host", "localhost")
        port = kwargs.get("port", 6379)
        self.redis_client = redis.StrictRedis(host=host, port=port, db=1)

    def get_client(self):
        return self.redis_client



import os

from bamboo_lib.models import PipelineStep, LinearPipelineExecutor
from bamboo_lib.steps import LoadStep
import pandas as pd
from bamboo_lib.connectors.models import Connector

from acs_core.fetch import Fetcher
from acs_core.pipeline import CubeGenerator

import yaml

# These are the geo hierarchy identifiers used to make api call.
# copied from fetch?
ZIP = "zip code tabulation area"
PLACE = 'place'
TRACT = 'tract'
COUNTY = 'county'
STATE = 'state'
NATION = 'us'
MSA = 'metropolitan statistical area/micropolitan statistical area'  # (320)


class ExtractStep(PipelineStep):
    def run_step(self, prev, params):
        year = params['year']
        sumlevels = params['sumlevels']
        estimate = params['estimate']

        pipeline_config_path = params['pipeline_config_path']
        fetcher = Fetcher(pipeline_config_path)
        return fetcher.fetch(
            years=[year],
            sumlevels=[sumlevels],
            estimates=[estimate])


class TransformStep(PipelineStep):
    def run_step(self, prev, params):
        dfs = prev
        # temp, TODO remove
        pipeline_config = yaml.load(open(params['pipeline_config_path']))
        print(pipeline_config)
        cube_generator = CubeGenerator(
            pipeline_config=pipeline_config,
            cube_name=pipeline_config['name'],
            value_label=pipeline_config['acs_table']['value_label'],
            input_dfs=dfs,
            schema='acs')
        return cube_generator.run()


def run(params, **kwargs):
    BAMBOO_LIB_DIR = os.environ.get("BAMBOO_LIB_DIR")
    conn_path = os.path.join(BAMBOO_LIB_DIR, "example", "conns.yaml")
    connector = Connector.fetch("postgres-local", open(conn_path))
    pipeline_config_path = params.get("pipeline_config_path")
    config_obj = yaml.load(open(pipeline_config_path))
    step1 = ExtractStep()
    step2 = TransformStep()
    step3 = LoadStep(config_obj["name"], connector, schema="acs")

    pipeline = LinearPipelineExecutor([step1, step2, step3], params)
    pipeline.run_pipeline()


if __name__ == '__main__':
    BAMBOO_LIB_DIR = os.environ.get("BAMBOO_LIB_DIR")
    run({"year": 2012,
         "estimate": 1,
         "sumlevels": "zip code tabulation area",
         "write-to-db": True,
         "pipeline_config_path": os.path.join(BAMBOO_LIB_DIR, "bamboo_lib/pipelines/bamboo_acs_pipeline/config/sex_by_age.yaml")})



from bamboo_lib.models import Parameter, BasePipeline
import acs_bamboo.standard_acs as standard_acs

from acs_bamboo.standard_acs import ZIP, PLACE, TRACT, COUNTY, STATE, NATION, MSA
from os.path import dirname, abspath, join
from os import listdir
SUM_LEVELS = [ZIP, PLACE, TRACT, COUNTY, STATE, NATION, MSA]


def get_parent_dir():
    parent_dir = dirname(dirname(abspath(__file__)))
    return join(parent_dir, "config")


def get_configs():
    parent_dir = get_parent_dir()
    configs = [join(parent_dir, f) for f in listdir(parent_dir)]
    return configs


class AcsPipeline(BasePipeline):

    @staticmethod
    def pipeline_id():
        return 'bamboo-acs-pipeline'

    @staticmethod
    def name():
        return 'ACS Pipeline'

    @staticmethod
    def description():
        return 'Processes data from ACS Summary Files'

    @staticmethod
    def website():
        return 'https://www.census.gov/programs-surveys/acs/data.html'

    @staticmethod
    def parameter_list():
        source_list = get_configs()
        return [
            Parameter(label="Source", name="pipeline_config_path", dtype=str, options=source_list),
            Parameter(name="year", dtype=int, allow_multiple=True,
                      options=range(2009, 2017)),
            Parameter(name="sumlevels", options=SUM_LEVELS, dtype=str),
            Parameter(name="estimate", dtype=int, options=[1, 5]),
            Parameter(label="Save results in database", name="write-to-db", dtype=bool, options=[True, False])
        ]

    @staticmethod
    def run(params_dict, **kwargs):
        print(params_dict)
        return standard_acs.run(params_dict)

    @staticmethod
    def get_sources():
        print(get_configs())



import re
import zipfile

from bamboo_lib.helpers import grab_connector, random_char
from bamboo_lib.models import PipelineStep, ResultWrapper
from bamboo_lib.logger import logger

class DownloadStep(PipelineStep):
    def __init__(self, connector=None, connector_path=None, callback=None, force=False):
        if not connector:
            raise ValueError("Must specify a connector")
        if isinstance(connector, str):
            cpath = connector_path or __file__
            self.connector = grab_connector(cpath, connector)
        else:
            self.connector = connector
        self.callback = callback
        self.force = force

    def run_step(self, prev, params):
        logger.debug("DOWNLOAD STEP")
        # Download Cube (URL defined in conns.yaml file.)
        # This step will save the contents of the connector target
        # to a file and then pass a file path to the next step
        return self.connector.download(params=params, callback=self.callback, force=self.force)

class LoadStep(PipelineStep):
    OPTIONS = ["if_exists", "schema", "index", "index_label", "chunksize", "dtype", "pk", "table_schema_only",
        "data_only", "connector_path", "nullable_list", "engine", "engine_params"]

    def __init__(self, table_name, connector, **kwargs):
        self.table_name = table_name
        if isinstance(connector, str):
            cpath = kwargs.get("connector_path", None) or __file__
            self.connector = grab_connector(cpath, connector)
            if "connector_path" in kwargs:
                del kwargs["connector_path"]
        else:
            self.connector = connector
        for key, val in kwargs.items():
            if key in LoadStep.OPTIONS:
                setattr(self, key, val)
            else:
                raise ValueError("Invalid parameter", key, val)

    def run_step(self, prev, params):
        logger.info("Running LoadStep step...")
        df = prev
        kwargs = {key: getattr(self, key) for key in self.OPTIONS if hasattr(self, key)}
        self.connector.write_df(self.table_name, df, **kwargs)
        return prev


class LoadStepDynamic(PipelineStep):
    OPTIONS = ["if_exists", "schema", "index", "index_label", "chunksize", "dtype", "pk"]

    def __init__(self, df_key, table_name_key, connector, **kwargs):
        self.df_key = df_key
        self.table_name_key = table_name_key
        self.connector = connector
        for key, val in kwargs.items():
            if key in LoadStep.OPTIONS:
                setattr(self, key, val)
            else:
                raise ValueError("Invalid parameter", key, val)

    def run_step(self, prev, params):
        logger.info("Running LoadStep step...")
        df = prev.get(self.df_key)
        kwargs = {key: getattr(self, key) for key in self.OPTIONS if hasattr(self, key)}
        self.connector.write_df(prev.get(self.table_name_key), df, **kwargs)
        return prev


class UnzipStep(PipelineStep):
    def __init__(self, compression='zip', pattern=None):
        self.compression = compression
        supported_compression = ['zip']
        self.pattern = pattern
        if self.compression not in supported_compression:
            raise Exception("extension not supported!")

    def run_step(self, filepath, params):
        if self.compression == 'zip':
            zfile = zipfile.ZipFile(filepath)
            for finfo in zfile.infolist():
                if not self.pattern or re.search(self.pattern, finfo.filename):
                    yield zfile.open(finfo)

        # return compressor(full_path)


class WriteDFToDiskStep(PipelineStep):
    def __init__(self, target_path, compression=None):
        self.target_path = target_path
        self.compression = compression

    def run_step(self, df, params):
        df.to_csv(self.target_path, index=False, compression=self.compression)
        return self.target_path


class SCPTransferStep(PipelineStep):
    def __init__(self, target_path, connector, **kwargs):
        if not target_path or not connector:
            raise Exception("You must specify a target path and a connector")
        super(SCPTransferStep, self).__init__(**kwargs)
        self.target_path = target_path
        self.connector = connector

    def run_step(self, file_obj, params):
        logger.debug("Transfering schema file: {} ...".format(self.target_path))
        use_fo = not isinstance(file_obj, str)
        self.connector.send_file(file_obj, self.target_path, use_fo=use_fo)
        logger.debug("Transfer complete!")
        return True


class SSHCommandStep(PipelineStep):
    def __init__(self, cmd, connector, **kwargs):
        if not cmd or not connector:
            raise Exception("You must specify a target path and a connector")
        super(SSHCommandStep, self).__init__(**kwargs)
        self.cmd = cmd
        self.connector = connector

    def run_step(self, input, params):
        logger.debug("Running command on remote host: {} ...".format(self.cmd))
        # self.connector.send_file(file_obj, self.target_path, use_fo=True)
        output = self.connector.run_command(self.cmd)
        logger.debug("Command complete! Result was:\n\n {}\n\n".format(output))
        return ResultWrapper(previous_result=input, current_result=output)


class LockStep(PipelineStep):
    def __init__(self, lock_name, redis_connector, next_step):
        import sherlock  # -- only use if needed
        sherlock.configure(timeout=45, backend=sherlock.backends.REDIS, client=redis_connector.get_client())
        self.lock_name = lock_name
        self.lock = sherlock.Lock(lock_name)
        self.next_step = next_step

    def run_step(self, prev, params):
        logger.debug("Waiting for lock {} ...".format(self.lock_name))
        self.lock.acquire()
        logger.debug("Acquired lock {} ...".format(self.lock_name))
        logger.debug("Running step...")
        result = self.next_step.run_step(prev, params)
        self.lock.release()
        logger.debug("Lock released!")
        return result


class SSHTunnelStartStep(PipelineStep):
    def __init__(self, sshtunnel_connector):
        self.sshtunnel_connector = sshtunnel_connector

    def run_step(self, prev, params):
        self.sshtunnel_connector.start()
        return prev


class SSHTunnelCloseStep(PipelineStep):
    def __init__(self, sshtunnel_connector):
        self.sshtunnel_connector = sshtunnel_connector

    def run_step(self, prev, params):
        self.sshtunnel_connector.close()
        return prev


class IngestMonetStep(PipelineStep):
    def __init__(self, table_name, schema, db_name, conns_path, compression=None):
        self.table_name = table_name
        self.schema = schema
        self.db_name = db_name
        self.conns_path = conns_path
        self.compression = compression

    def run_step(self, prev, params):
        # connectors config
        server_str = params.get("server-connector")
        sftp_connector = grab_connector(self.conns_path, server_str)

        monet_connector = grab_connector(self.conns_path, "monet-remote")
        redis_connector = grab_connector(self.conns_path, "redis-remote")

        lock_name = params.get("lock_name", "monet-lock")

        # prep csv file for write to disk and transfer to server
        # (including compression)
        random_filename = random_char(32)
        target_path = "/tmp/{}-{}.csv".format(self.schema, random_filename)
        if self.compression == "gzip":
            target_path = target_path + ".gz"
        elif self.compression == "bz2":
            target_path = target_path + ".bz2"

        write_to_disk_step = WriteDFToDiskStep(target_path=target_path, compression=self.compression)
        transfer_step = SCPTransferStep(target_path, sftp_connector)

        # Create table
        # Must use lock because Monetdb does not support table creation concurrent
        # with any other transaction (including COPY)
        # put the table gen into a load step which is wrapped by the lock step
        # for ssh tunnel, just have to start one manually, and then latch onto it with
        # the ssh_tunnel step by specifying the port.
        # e.g. ssh -L 6379:localhost:6379 deploy@canon -N

        create_table_step = LoadStep(self.table_name, monet_connector, table_schema_only=True, schema=self.schema)
        create_table_with_lock_step = LockStep(lock_name, redis_connector, create_table_step)

        # Ingest
        ingest_cmd = '''mclient -d {} -h localhost -s "COPY OFFSET 2 INTO {}.{} FROM '{}' USING DELIMITERS ',', '\n', '\\\"' NULL AS '' "'''.format(self.db_name, self.schema, self.table_name, target_path)
        remote_cmd_step = SSHCommandStep(ingest_cmd, sftp_connector)

        # Run steps
        res_1 = create_table_with_lock_step.run_step(prev, params)
        res_2 = write_to_disk_step.run_step(res_1, params)
        res_3 = transfer_step.run_step(res_2, params)
        res_4 = remote_cmd_step.run_step(res_3, params)

        return res_4


class BaseGenerateSchemaStep(PipelineStep):
    """ This step serves as a base for generating dynamic schemas. It consists
    of helper methods to generate elements to populate template XML files which
    are then output into a <cube>.xml in a desired location.
    The following is an outline of the basic structure currently supported:

    - Cube
      - Annotation(s)
      - Dimension(s)
        - Annotation(s)
        - Hierarchy(s)
          - Table
          - Level(s)
      - Measure(s)
    """

    @staticmethod
    def create_annotation(name, value):
        """ Creates an Annotation object. After passing through the template
        builder, the output will look like:

        <Annotation name="title">Title</Annotation>

        :param name: name of this annotation
        :param value: value for this annotation
        :return: Annotation object
        """
        return {
            "name": name,
            "value": value
        }

    @staticmethod
    def create_dimension(name, hierarchies, foreign_key="", annotations=None):
        """ Creates a Dimension object. After passing through the template
        builder, the output will look like:

        <Dimension foreignKey="region_id" name="Region">
            ... Annotations and Hierarchies can go in here ...
        </Dimension>

        :param name: name of this dimension
        :param hierarchies: list of hierarchies for this dimension
        :param foreign_key: name of column in the fact table used to connect to
                            a dimension table
        :param annotations: list of annotations
        :return: Dimension object
        """
        if not isinstance(hierarchies, list):
            raise TypeError("'hierarchies' should be a list")
        if annotations and not isinstance(annotations, list):
            raise TypeError("'annotations' should be a list")

        return {
            "name": name, "foreign_key": foreign_key, "annotations": annotations,
            "hierarchies": hierarchies
        }

    @staticmethod
    def create_hierarchy(table, levels, name="", primary_key="", has_all=""):
        """ Creates a Hierarchy object. After passing through the template
        builder, the output will look like:

        <Hierarchy name="City" primaryKey="geoid" hasAll="true">
            ... Table and Levels will be rendered here ...
        </Hierarchy>

        :param table: dimension table for this hierarchy
        :param levels: levels of this hierarchy
        :param name: name of this hierarchy
        :param primary_key: primary key in the dimension table
        :param has_all: whether the hierarchy should have an 'All' level
        :return: Hierarchy object
        """
        if not isinstance(table, dict):
            raise TypeError("'table' should be a dict")
        if not isinstance(levels, list):
            raise TypeError("'levels' should be a list")

        return {
            "table": table, "levels": levels, "name": name,
            "has_all": has_all, "primary_key": primary_key,
        }

    @staticmethod
    def create_table(name, schema):
        """ Creates a Table object. After passing through the template
        builder, the output will look like:

        <Table name="sales" schema="store_sales" />

        :param name: name of this table
        :param schema: schema where the table is located
        :return: Table object
        """
        return {"name": name, "schema": schema}

    @staticmethod
    def create_level(name, column, _type="", level_type="", name_column="",
                     unique_members=""):
        """ Creates a Level object. After passing through the template
        builder, the output will look like:

        <Level name="Year" column="year" type="Numeric" levelType="TimeYears"
               nameColumn="year_name" uniqueMembers="true" />

        :param name: name of this level
        :param column: column in the dimension table corresponding to this level
        :param _type:
        :param level_type:
        :param name_column: name column in the dimension table for this level
        :param unique_members: whether the members of this
        :return: Level object
        """
        return {
            "name": name, "column": column, "type": _type,
            "level_type": level_type, "name_column": name_column,
            "unique_members": unique_members
        }

    @staticmethod
    def create_measure(name, column, aggregator):
        """ Creates a Measure object. After passing through the template
        builder, the output will look like:

        <Measure name="Imports" column="imports" aggregator="sum" />

        :param name: name of this measure
        :param column: column in the fact table corresponding to this measure
        :param aggregator: aggregator for this measure
        :return: Measure object
        """
        return {"name": name, "column": column, "aggregator": aggregator}

    @staticmethod
    def write_cube_file(cube_name, table_name, schema_name, annotations,
                        dimensions, measures, file_location, file_name):
        """ Writes a cube XML file into a desired location using the `cube.xml`
        template.

        :param cube_name: name of the cube being created
        :param table_name: name of the fact table for this cube
        :param schema_name: name of the schema where the fact table is located
        :param annotations: list of annotations
        :param dimensions: list of dimensions
        :param measures: list of measures
        :param file_location: folder where the file should be written
        :param file_name: name of the file
        :return template: the template generated
        """
        env = Environment(
            loader=PackageLoader("bamboo_lib", "templates"),
            autoescape=select_autoescape(["xml"])
        )

        template = env.get_template("cube.xml")
        template = template.render(
            cube_name=cube_name,
            table_name=table_name,
            schema_name=schema_name,
            annotations=annotations,
            dimensions=dimensions,
            measures=measures
        )

        with open("%s/%s.xml" % (file_location, file_name), "w") as f:
            f.write(template)

        return template



import os
import pandas as pd
import numpy as np
from numpy.random import choice

COL_RATE = "Total Conversion Rate"
SOC_00 = "socp00"
SOC_10 = "socp10"
SOC_12 = "socp12"

MALE_VAL = 1
FEMALE_VAL = 2

DEGREE = 'schl'
SEX = 'sex'
ESTIMATE = 'estimate'
YEAR = 'year'

data_dir = os.path.dirname(__file__)


def get_path(target):
    return os.path.join(data_dir, target)


def index(df, idx):
    return df.set_index(idx)


soc_direct_map = pd.read_csv(get_path('data/SOC_00_to_10_direct.csv'), converters={"socp00": str, "socp10": str})
soc_direct_map = soc_direct_map.to_dict(orient="records")
soc_direct_map = {x[SOC_00]: x[SOC_10] for x in soc_direct_map}

soc12_direct_map = pd.read_csv(get_path('data/SOC_10_to_12_direct.csv'), converters={"socp12": str, "socp10": str})
soc12_direct_map = soc12_direct_map.to_dict(orient="records")
soc12_direct_map = {x[SOC_10]: x[SOC_12] for x in soc12_direct_map}

soc_hs_m_map = pd.read_csv(get_path('data/SOC_00_to_10_HS_M.csv'), converters={"socp00": str, "socp10": str})
soc_hs_f_map = pd.read_csv(get_path('data/SOC_00_to_10_HS_F.csv'), converters={"socp00": str, "socp10": str})
soc_ba_m_map = pd.read_csv(get_path('data/SOC_00_to_10_BA_M.csv'), converters={"socp00": str, "socp10": str})
soc_ba_f_map = pd.read_csv(get_path('data/SOC_00_to_10_BA_F.csv'), converters={"socp00": str, "socp10": str})
soc_adv_m_map = pd.read_csv(get_path('data/SOC_00_to_10_ADV_M.csv'), converters={"socp00": str, "socp10": str})
soc_adv_f_map = pd.read_csv(get_path('data/SOC_00_to_10_ADV_F.csv'), converters={"socp00": str, "socp10": str})
to_index = [soc_hs_m_map, soc_hs_f_map, soc_ba_m_map, soc_ba_f_map, soc_adv_m_map, soc_adv_f_map]
soc_hs_m_map, soc_hs_f_map, soc_ba_m_map, soc_ba_f_map, soc_adv_m_map, soc_adv_f_map = map(lambda x: index(x, SOC_00), to_index)

soc12_hs_m_map = pd.read_csv(get_path('data/SOC_10_to_12_HS_M.csv'), converters={"socp12": str, "socp10": str})
soc12_hs_f_map = pd.read_csv(get_path('data/SOC_10_to_12_HS_F.csv'), converters={"socp12": str, "socp10": str})
soc12_ba_m_map = pd.read_csv(get_path('data/SOC_10_to_12_BA_M.csv'), converters={"socp12": str, "socp10": str})
soc12_ba_f_map = pd.read_csv(get_path('data/SOC_10_to_12_BA_F.csv'), converters={"socp12": str, "socp10": str})
soc12_adv_m_map = pd.read_csv(get_path('data/SOC_10_to_12_ADV_M.csv'), converters={"socp12": str, "socp10": str})
soc12_adv_f_map = pd.read_csv(get_path('data/SOC_10_to_12_ADV_F.csv'), converters={"socp12": str, "socp10": str})
to_index = [soc12_hs_m_map, soc12_hs_f_map, soc12_ba_m_map, soc12_ba_f_map, soc12_adv_m_map, soc12_adv_f_map]
soc12_hs_m_map, soc12_hs_f_map, soc12_ba_m_map, soc12_ba_f_map, soc12_adv_m_map, soc12_adv_f_map = map(lambda x: index(x, SOC_10), to_index)

xforms = {SOC_00: SOC_10, SOC_10: SOC_12}


def get_soc_mode(var_map):
    year, est = map(int, [var_map[YEAR], var_map[ESTIMATE]])
    if year >= 2012:
        return SOC_12
    elif year >= 2009:
        return SOC_10
    else:
        return SOC_00


def is_old_school(var_map):
    return (int(var_map[YEAR]) - int(var_map[ESTIMATE]) + 1) <= 2007


def randomizer(code, rule_map, start_col):
    if code not in rule_map.index:
        return code

    tmpdf = rule_map.ix[code]

    if tmpdf.empty:
        return code

    if isinstance(tmpdf, pd.Series):
        return tmpdf[xforms[start_col]]

    idx = choice(range(len(tmpdf)), p=tmpdf[COL_RATE].values)

    return tmpdf.iloc[idx][xforms[start_col]]


def _convert(df, start_col, is_oldschool_mode):
    ''' Use conversion tables to transform across classifications '''
    if is_oldschool_mode:
        raise NotImplementedError("Requires additional processing...")
    HS_VAL, BA_VAL, ADV_VAL = [20, 21, 22]

    if start_col not in df.columns:
        raise Exception(start_col, "Not in", df.columns)
        return df

    if start_col == SOC_10:
        direct_map = soc12_direct_map
        hs_m_map = soc12_hs_m_map
        hs_f_map = soc12_hs_f_map
        ba_m_map = soc12_ba_m_map
        ba_f_map = soc12_ba_f_map
        adv_m_map = soc12_adv_m_map
        adv_f_map = soc12_adv_f_map
    elif start_col == SOC_00:
        direct_map = soc_direct_map
        hs_m_map = soc_hs_m_map
        hs_f_map = soc_hs_f_map
        ba_m_map = soc_ba_m_map
        ba_f_map = soc_ba_f_map
        adv_m_map = soc_adv_m_map
        adv_f_map = soc_adv_f_map
    else:
        raise Exception("BAD START COLUMN")
    # -- First apply direct transformation then split into groups
    tmpcond = df[start_col].isin(direct_map.keys())
    df.loc[tmpcond, start_col] = df[tmpcond][start_col].map(direct_map)
    # df.loc[df[start_col].isin(direct_map.keys()), start_col] = df[start_col].map(direct_map)

    # print(direct_map)
    # print(df.head())
    # print(start_col)
    # print(df[start_col])
    # print(df[start_col].isin(direct_map.keys()))

    # print("HEL" , df.loc[df[start_col].isin(direct_map.keys())])
    # -- Intelligently split the dataframe into six parts based on gender & edu
    HS = (df[DEGREE].astype(float) <= HS_VAL)
    BA = (df[DEGREE].astype(float) == BA_VAL)
    ADV = (df[DEGREE].astype(float) >= ADV_VAL)
    MALE = (df[SEX].astype(int) == MALE_VAL)
    FEMALE = (df[SEX].astype(int) == FEMALE_VAL)

    HS_MALE = HS & MALE
    HS_FEMALE = HS & FEMALE
    BA_MALE = BA & MALE
    BA_FEMALE = BA & FEMALE
    ADV_MALE = ADV & MALE
    ADV_FEMALE = ADV & FEMALE
    IS_VALID_ESR = df.esr.notnull()
    EVERYTHING_ELSE = (~HS_MALE & ~HS_FEMALE & ~BA_MALE & ~BA_FEMALE & ~ADV_MALE & ~ADV_FEMALE) & IS_VALID_ESR

    if not df[EVERYTHING_ELSE][start_col].empty:
        print(df[EVERYTHING_ELSE])
        print(len(df[EVERYTHING_ELSE]))
        print(df[EVERYTHING_ELSE].schl.unique())
        raise Exception("*** ERROR! Unaccounted for people")

    rules = [(HS_MALE, hs_m_map), (HS_FEMALE, hs_f_map), (BA_MALE, ba_m_map), (BA_FEMALE, ba_f_map), (ADV_MALE, adv_m_map), (ADV_FEMALE, adv_f_map)]
    rule_num = 0
    for (rule, rule_map) in rules:
        print("Applying rule {},".format(rule_num))
        rule = rule & df[start_col].notnull()
        df.loc[rule & IS_VALID_ESR, xforms[start_col]] = df.loc[rule & IS_VALID_ESR, start_col].apply(lambda x: randomizer(x, rule_map, start_col))
        rule_num += 1
    return df


def occ_convert(df, var_map):
    '''
    PUMS occupational crosswalking
    There is data in PUMS 5-year files with both SOCP00 and SOCP10 formats.
    To deal with this, we must first crosswalk SOCP00 codes to SOCP10 codes,
    then we take all SOCP10 codes and walk them over to SOCP12 codes. So We end
    up with a SOCP12 column with the crosswalked codes.
    '''
    is_oldschool_mode = is_old_school(var_map)
    soc_mode = get_soc_mode(var_map)

    if SOC_00 in df.columns:
        print("Spotted socp00 in columns...first convert this.")

        df = _convert(df, SOC_00, is_oldschool_mode)
        df = _convert(df, SOC_10, is_oldschool_mode)
    elif SOC_10 in df.columns:
        print("Spotted socp10 in columns...first convert this.")
        print(df, "HERE!!!")
        df = _convert(df, SOC_10, is_oldschool_mode)
    else:
        print("*** single socp ****")
        print("RENAMING TO", soc_mode)
        print("SOC cols 1:", [x for x in df.columns if "soc" in x.lower()])
        df.rename(columns={"socp": soc_mode}, inplace=True)
        print("SOC cols 2:", [x for x in df.columns if "soc" in x.lower()])

        if soc_mode == SOC_00:
            print("****** PART I")
            df = _convert(df, SOC_00, is_oldschool_mode)
            print("****** PART II")
            df = _convert(df, SOC_10, is_oldschool_mode)
        elif soc_mode == SOC_10:
            df = _convert(df, SOC_10, is_oldschool_mode)
        elif soc_mode == SOC_12:
            pass  # Nothing to do!

    df.rename(columns={SOC_12: "socp"}, inplace=True)
    return df


if __name__ == '__main__':
    moi = pd.DataFrame({"x": [100, 44], "socp00": [np.nan, "472020"], "socp10": ["472XXX", np.nan], "sex": [2, 2], "degree": [10, 10]})
    print("Starting with")
    print(moi)
    print("####")
    res = occ_convert(moi, {YEAR: 2006, ESTIMATE: 3})
    print("CONVERTED TO")
    print(res)



import sqlalchemy
from bamboo_lib.connectors.models import BaseDriver
from data_catapult.database.postgres import PostgresDriver as CatapultPgd


class PostgresDriver(BaseDriver):
    """
    Driver that allows bulk data exports to Postgres databases.

    Example usage:

    >>> import pandas as pd
    >>> df = pd.DataFrame({"x": [1, 2, 3, 4]})
    >>> pgd = PostgresDriver(uri="postgres://user:password@localhost/your_db")
    >>> pgd.write_df("my_new_table", df, schema="my_schema")
    """
    TYPE = 'Postgres Bulk Write Driver'

    def __init__(self, **kwargs):
        super(PostgresDriver, self).__init__(**kwargs)
        self.engine = sqlalchemy.create_engine(self.uri)

    def write_df(self, table_name, df, **kwargs):
        pgd = CatapultPgd(self.engine)
        schema_name = kwargs.get("schema", "public")
        dtype = kwargs.get("dtype", None)
        return pgd.to_sql(df, schema_name, table_name, dtype)

    def add_pk(self, *args):
        pgd = CatapultPgd(self.engine)
        return pgd.add_pk(*args)



import os
import sqlalchemy
import yaml
import importlib
import re
import ast


class BaseDriver(object):
    """
    BaseDriver represents an abstract base class for data source drivers.

    The constructor of BaseDriver takes any number of keyword arguments (kwargs) that it will store internally.
    """
    TYPE = 'Generic DB Driver'

    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)
        self.uri = kwargs["uri"] if "uri" in kwargs else ""
        self._internal_data = kwargs

    @classmethod
    def from_json(cls, raw_setts):
        """Instantiates an instance of the BaseDriver based on the raw_setts provided.

        :param raw_setts: a string containing the keyword arguments for the driver.

        :returns: An `object` of the target BaseDriver with specified settings.
        """
        setts = ast.literal_eval(raw_setts)
        return cls(**setts)

    @classmethod
    def get_name(cls):
        return cls.TYPE

    @staticmethod
    def resolve_params(input_str, params_dict):
        matches = re.findall('<([A-Za-z0-9_ ]+)>', input_str)
        for match in matches:
            input_str = input_str.replace("<{}>".format(match), str(params_dict[match]))
        return input_str

    @classmethod
    def get_module_path(cls):
        return cls.__module__ + "." + cls.__name__


class Connector(object):
    """
    The Connector object represents the metadata about a driver and
    its desired settings so that they may be serialized and deserialized as needed.
    """
    @staticmethod
    def fetch(name, connections_source):
        if isinstance(connections_source, str):
            source_dict = ast.literal_eval(connections_source)
        else:
            source_dict = yaml.load(connections_source, Loader=yaml.FullLoader)
            if name not in source_dict:
                raise ValueError("Invalid Connection name", name)
            source_dict = source_dict[name]

        if "driver" not in source_dict:
            raise ValueError("Source must contain a driver value", source_dict)

        source_dict = {key: os.path.expandvars(val) if isinstance(val, str) else val for key, val in source_dict.items()}
        driver_str = source_dict["driver"]
        # resolve driver
        clzz_file, clzz_name = driver_str.rsplit(".", 1)
        clzz_file = importlib.import_module(clzz_file)
        if not hasattr(clzz_file, clzz_name):
            raise ValueError("Class file", clzz_file, "does not contain a class called", clzz_name)
        driver_obj = getattr(clzz_file, clzz_name)
        if "connection_parameters" not in source_dict:
            driver_obj = driver_obj(**source_dict)
        else:
            driver_obj = driver_obj(**source_dict["connection_parameters"])
        return driver_obj


class DatabaseDriver(BaseDriver):
    """Generic database driver that is a passthrough to sqlalchemy's create_engine"""
    def __init__(self, **kwargs):
        super(DatabaseDriver, self).__init__(**kwargs)
        self.engine = sqlalchemy.create_engine(self.uri)

    def write_df(self, table_name, df, **kwargs):
        return df.to_sql(table_name,
                         self.engine,
                         **kwargs)



"""
.. module:: models
   :synopsis: Core data structure models
"""
import abc
from bamboo_lib.logger import logger
from bamboo_lib.helpers import random_char

class BasePipeline(object):
    """
    BasePipeline is the abstract base class that all pipelines must extend.

    .. note::

       You must override the methods of:
         * :func:`~BasePipeline.pipeline_id`.
         * :func:`~BasePipeline.name`
         * :func:`~BasePipeline.description`
         * :func:`~BasePipeline.website`
         * :func:`~BasePipeline.run`
    """

    __metaclass__ = abc.ABCMeta

    @staticmethod
    @abc.abstractmethod
    def validate_params(params):
        return True

    @staticmethod
    @abc.abstractmethod
    def pipeline_id():
        """
        :returns: `str` that uniquely identifies the pipeline.
        """
        raise NotImplementedError("Please implement the pipeline_id() method")

    @staticmethod
    @abc.abstractmethod
    def name():
        """
        :returns: `str` a human-readable name describing the pipeline.
        """
        raise NotImplementedError("Please implement the name() method")

    @staticmethod
    @abc.abstractmethod
    def description():
        """
        :returns: `str` description of the pipeline.
        """
        raise NotImplementedError("Please implement the description() method")

    @staticmethod
    @abc.abstractmethod
    def params():
        """
        :returns: `list` containing :class:`.Parameter` objects.
        """
        raise NotImplementedError("Please implement the params() method")

    @staticmethod
    @abc.abstractmethod
    def website():
        raise NotImplementedError("Please implement the website() method")

    @staticmethod
    def run(params_dict, **kwargs):
        raise NotImplementedError("Please implement the run() method")

class EasyPipeline(BasePipeline):
    """
    EasyPipeline is a simple implementation of the BasePipeline class to reduce the boilerplate
    needed for setting up a new Bamboo pipeline.

    The only method strictly required is the steps static method. Each implementation of EasyPipeline should
    create a steps method that returns a list of `PipelineStep`. If your pipeline accepts parameters,
    you will also need to override the :func:`~EasyPipeline.parameter_list` static class method which returns
    a list of acceptable parameters for the pipeline.

    By default, EasyPipeline only supports linear pipelines (i.e. those without foreach loops). Though you may override the
    :func:`~EasyPipeline.run` method to customize this behavior if needed in the meantime.

    .. note::

       You must override the method:
         * :func:`~EasyPipeline.steps`.

    .. code-block:: python

        from bamboo_lib.models import Parameter, EasyPipeline, PipelineStep
        from bamboo_lib.steps import DownloadStep, LoadStep

        class MySimplePipeline(EasyPipeline):
            @staticmethod
            def parameter_list():
                return [Parameter("year", dtype=int)]

            @staticmethod
            def steps():
                dl_step = DownloadStep(connector="my-data", connector_path=__file__)
                xform_step = TransformStep()
                load_step = LoadStep()
                return [dl_step, xform_step, load_step]
    """
    @classmethod
    def name(cls):
        return cls.__name__

    @staticmethod
    def pipeline_id():
        return EasyPipeline.name()

    @staticmethod
    def parameter_list():
        """
        :returns: `list` of `Parameter` objects that define the possible inputs to the pipeline.
        """
        return []

    @staticmethod
    def description():
        return ""

    @staticmethod
    def website():
        return "N/A"

    @staticmethod
    @abc.abstractmethod
    def steps(params):
        """
        :param params: The original parameters passed into the pipeline.
        :returns: `list` of `PipelineStep` objects that define the steps to be executed for the pipeline.
        """
        raise NotImplementedError("Please make sure your subclass of EasyPipeline implements the steps() function")

    @classmethod
    def run(cls, params, **kwargs):
        pipeline = AdvancedPipelineExecutor(params)
        for step in cls.steps(params):
            pipeline = pipeline.next(step)
        return pipeline.run_pipeline()

class Parameter(object):
    """
    A Parameter represents an input to a pipeline. In general, any value for a pipeline that
    should be user-configurable should be a parameter.

      * name is `str` describing the pipeline.
      * dtype is an `object` representing the type of the parameter (e.g. `int`, `str`).
      * options represents a `list` of values that are valid for the parameter.
      * allow_multiple is a `bool` whether in a user-interface users can select multiple values for this parameter.
      * source represents a `str` huh?
    """
    def __init__(self, name, dtype, options=None, allow_multiple=False, label="", source=None):
        """

        """
        self.name = name
        self.dtype = dtype
        self.options = options
        self.allow_multiple = allow_multiple
        self.label = label
        self.source = source


class LinearPipelineExecutor(object):
    def __init__(self, steps=None, params=None):
        self.steps = steps
        self.params = params

    def set_steps(self, steps):
        self.steps = steps

    # Run step, save output, move next
    def run_pipeline(self):
        prev_result = None
        # have pipeline catch then rethrow error
        logger.info("==== Pipeline run started...")
        for step in self.steps:
            logger.info("Starting step %s ....", str(step.__class__))
            result = step.run_step(prev_result, self.params)
            logger.info("Ending step %s ....", str(step.__class__))
            prev_result = result
        logger.info("==== Pipeline run completed!")


class GraphPipelineExecutor(object):
    def __init__(self, for_each=None, do_steps=None, params=None):
        self.for_each = for_each
        self.do_steps = do_steps
        self.params = params

    # Run step, save output, move next
    def run_pipeline(self):
        logger.info("==== Pipeline run started...")
        count = 1
        for chunk_result in self.for_each.run_step(None, self.params):
            logger.info("Starting pass {}".format(count))
            prev_result = chunk_result
            for step in self.do_steps:
                logger.info("Starting step %s ....", str(step.__class__))
                result = step.run_step(prev_result, self.params)
                logger.info("Ending step %s ....", str(step.__class__))
                prev_result = result
            count += 1
        logger.info("==== Pipeline run completed!")


class PipelineStep(object):
    """
    This class represents the basic unit of work in a pipeline. The outputs
    from steps are chained together, such that the output of step 1 is the input to step 2 and so on.

    Any keyword arguments passed to the `PipelineStep` constructor will be set as
    attributes in the object.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)

    @abc.abstractmethod
    def run_step(self, prev_result, params):
        """
        :param prev_result: The output of the previous step in the pipeline.
        :param params: The original parameters passed into the pipeline.

        :returns: An `object` that is the result of the processing of this step.
        """


class Node(object):
    def __init__(self, step, is_iterator=False):
        self.step = step
        self.next = None
        self.prev = None
        self.children = None
        self.is_iterator = is_iterator
        self.iterator = None
        self.started = False


class EndNode(Node):
    def __init__(self):
        self.next = None
        self.prev = None
        self.is_iterator = False


class ComplexPipelineExecutor(object):
    def __init__(self, params=None):
        self.execution_plan = []
        self.params = params
        # self.pointer = None
        self.root = None
        self.pointer = None

    def insert(self, head, node_to_add):
        if not self.root and not head:
            self.root = node_to_add
        elif head.next:
            return self.insert(head.next, node_to_add)
        else:
            head.next = node_to_add
            node_to_add.prev = head
        self.pointer = node_to_add
        return node_to_add

    def next(self, step):
        node = Node(step)
        self.insert(self.root, node)
        return self

    def foreach(self, step):
        node = Node(step, is_iterator=True)
        self.insert(self.root, node)
        return self

    def endeach(self):
        node = EndNode()
        self.insert(self.root, node)
        return self

    def __str__(self):
        return str(self.execution_plan)

    def run_pipeline(self):
        return self.run_pipeline_helper(self.root, None, stack=[], history=[])

    def run_pipeline_helper(self, curr_node, prev_result, stack=None, history=None):
        if not curr_node:
            logger.debug("Done.")
            return prev_result
        elif curr_node.is_iterator and not curr_node.started:
            iterator = curr_node.step.run_step(prev_result, self.params)
            # print "ITERATOR CREATION", iterator
            stack.insert(0, [iterator, curr_node])
            curr_node.started = True
            return self.run_pipeline_helper(curr_node.next, next(iterator), stack=stack, history=history)
        elif curr_node.is_iterator and curr_node.started:
            try:
                # logger.debug("Refreshing iterator...")
                iterator, target_node = stack[0]
                # refresh children iterators
                hiterator, htarget_node = history.pop(0)
                hiterator = htarget_node.step.run_step(prev_result, self.params)
                stack.insert(0, [hiterator, htarget_node])
                return self.run_pipeline_helper(curr_node.next, next(hiterator), stack=stack, history=history)
            except StopIteration:
                raise Exception("TODO ... not yet implemented!")

        elif isinstance(curr_node, EndNode):

            iterator, target_node = stack[0]
            # logger.debug("Running ITERATOR" + str(iterator))
            try:
                return self.run_pipeline_helper(target_node.next, next(iterator), stack=stack, history=history)
            except StopIteration:
                old_node = stack.pop(0)
                history.insert(0, old_node)

                # logger.debug("ITERATOR EXHAUSTED " + str(iterator))
                if stack:
                    iterator, target_node = stack[0]
                    # if there's a parent iterator, go to that, otherwise
                    # continue on...
                return self.run_pipeline_helper(curr_node.next, prev_result, stack=stack, history=history)
        else:
            # simple movement
            prev_result = curr_node.step.run_step(prev_result, self.params)
            return self.run_pipeline_helper(curr_node.next, prev_result, stack=stack, history=history)


class AdvancedPipelineExecutor(object):
    """AdvancedPipelineExecutor is the standard pipeline runner that most applications should utilize.

    Examples:
        Pipelines are passed dictionaries with key-value pairs for their parameters. Depending on the pipeline's `param` function,
        these raw values will then get parsed and validated into structured values. To assemble the pipeline,
        create a new AdvancedPipelineExecutor object passing in the raw dictionary. Using `next` you can specify the order in which
        the steps will run. Note that execution does not start until `run_pipeline` is called.

        For more information on creating steps, see :class:`.PipelineStep`.

        .. code-block:: python

           pipeline = AdvancedPipelineExecutor(raw_dict)
           pipeline = pipeline.next(extract_step).next(transform_step).next(load_step)
           pipeline.run_pipeline()
    """
    def __init__(self, params=None):
        self.execution_plan = []
        self.params = params
        self.iterator_idx_stack = []

    def next(self, step):
        self.execution_plan.append(["standard", step])
        return self

    def foreach(self, step):
        self.iterator_idx_stack.insert(0, len(self.execution_plan))
        self.execution_plan.append(["iterator", step])
        return self

    def endeach(self):
        if len(self.iterator_idx_stack) > 0:
            #  help iterators
            last_iterator_idx = self.iterator_idx_stack.pop(0)
            op, my_it_step = self.execution_plan[last_iterator_idx]
            setattr(my_it_step, "endpoint", len(self.execution_plan))
            # raise Exception(my_it_step.endpoint)
            self.execution_plan.append(["enditerator", last_iterator_idx])
        else:
            raise ValueError("Cannot have end iterator without matching start iterator")
        return self

    def next_endstate(self, start):
        for i in range(start, len(self.execution_plan)):
            if self.execution_plan[i][0] == 'enditerator':
                return i
        raise Exception("TODO!")

    def run_pipeline(self):
        pointer = 0
        prev_result = None
        it_stack = []
        it_stack_positions = []
        while pointer < len(self.execution_plan):
            instruction, step = self.execution_plan[pointer]

            if instruction == "standard":
                # This is a simple step, so just run it and increment the pointer
                # raise Exception("TODO: jump past end of iterator")
                result = step.run_step(prev_result, self.params)
                prev_result = result
                pointer += 1
            elif instruction == "iterator":
                # This is the start of a loop, what we will want to do is,
                # run this step, and loop-over over steps
                result = step.run_step(prev_result, self.params)
                # it_stack.insert(0, enumerate(result))
                it_stack.insert(0, result)
                it_stack_positions.insert(0, pointer + 1)
                prev_result = next(result)  # sets up iterator so that it will be injected into prev_result of next step
                pointer += 1
            elif instruction == "enditerator":
                try:
                    curr_iter = it_stack[0]
                    iter_result = next(curr_iter)
                    prev_result = iter_result
                    # move pointer back to start position
                    pointer = it_stack_positions[0]
                except StopIteration:
                    # prev_result = None
                    pointer += 1
                    it_stack.pop(0)
                    it_stack_positions.pop(0)
            else:
                raise ValueError("Invalid instruction!", instruction)
        return prev_result


class ResultWrapper(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)



# test_capitalize.py
import pytest
import os
from bamboo_lib.models import PipelineStep, AdvancedPipelineExecutor
from bamboo_lib.helpers import grab_connector



class DisplayStep(PipelineStep):
    def run_step(self, prev_result, params):
        print("RESULT:", prev_result, "\n")
        return prev_result

class DownloadStep(PipelineStep):
    def run_step(self, prev, params):
        # Download Cube (URL defined in conns.yaml file.)
        # This step will save the contents of the connector target
        # to a file and then pass a file path to the next step
        return self.connector.download(params=params)


@pytest.fixture()
def pipeline():
    source_connector = grab_connector(__file__, "http-local")
    http_dl_step = DownloadStep(connector=source_connector)
    show_step = DisplayStep()
    pp = AdvancedPipelineExecutor({})
    pp = pp.next(http_dl_step).next(show_step)
    return pp


def test_dl_default_path(pipeline):
    res = pipeline.run_pipeline()
    assert res == "/tmp/999f2fa455aff231b7daa3680efdb03abd078fa1b009e38677c42d24"

def test_dl_custom_path1(pipeline):
    os.environ["BAMBOO_DOWNLOAD_FOLDER"] = "/tmp/a2/"
    res = pipeline.run_pipeline()
    assert res == "/tmp/a2/999f2fa455aff231b7daa3680efdb03abd078fa1b009e38677c42d24"
    del os.environ["BAMBOO_DOWNLOAD_FOLDER"]

def test_dl_custom_path2(pipeline):
    os.environ["BAMBOO_DOWNLOAD_FOLDER"] = "/tmp/a2"
    res = pipeline.run_pipeline()
    assert res == "/tmp/a2/999f2fa455aff231b7daa3680efdb03abd078fa1b009e38677c42d24"
    del os.environ["BAMBOO_DOWNLOAD_FOLDER"]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

import os
from pathlib import Path
pardir = Path(__file__).parent

with open(os.path.join(pardir, "bamboo_lib", "VERSION"), "r") as vfh:
    version = vfh.read()

setuptools.setup(
    author="Jonathan Speiser",
    author_email="jonathan@datawheel.us",
    classifiers=(
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ),
    description="Python ETL library",
    include_package_data=True,
    install_requires=[
        "pandas==0.24.2",
        "sqlalchemy==1.2.10",
        "data-catapult",
        "paramiko==2.4.1",
        "sshtunnel==0.1.4",
        "redis==2.10.6",
        "pytest"
    ],
    name="bamboo_lib",
    license='All Rights Reserved',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    version=version,
    url="https://github.com/Datawheel/bamboo-lib",
)



from bamboo_lib.models import BasePipeline
from bamboo_lib.models import PipelineStep, ComplexPipelineExecutor  # ,Connector


class DummyStep(PipelineStep):
    def run_step(self, prev, params):
        return prev


class DummyPipeline(BasePipeline):
    @staticmethod
    def pipeline_id():
        return 'dummy'

    @staticmethod
    def name():
        return 'Dummy Pipeline'

    @staticmethod
    def description():
        return 'For testing only'

    @staticmethod
    def website():
        return 'https://www.census.gov/programs-surveys/acs/data/pums.html'

    @staticmethod
    def parameter_list():
        return []

    @staticmethod
    def run(params_dict, **kwargs):
        pp = ComplexPipelineExecutor({})
        pp.next(DummyStep())
        return pp.run_pipeline()



from bamboo_lib.connectors.models import BaseDriver
from data_catapult.database.monetdb import MonetDBDriver as CatapultMdb
import monetdb.sql
from sqlalchemy.engine.url import make_url


class MonetDriver(BaseDriver):
    TYPE = 'MonetDB Bulk Write Driver'

    def __init__(self, **kwargs):
        super(MonetDriver, self).__init__(**kwargs)

        username = kwargs.get("username", "monetdb")
        password = kwargs.get("password", "monetdb")
        hostname = kwargs.get("hostname", "localhost")
        database = kwargs.get("database", "public")
        # port = int(kwargs.get("port", 50000)) # TODO accept port

        if kwargs.get("uri", None):
            db_uri = make_url(kwargs["uri"])
            username = db_uri.username
            hostname = db_uri.host
            password = db_uri.password
            database = db_uri.database

        self.conn = monetdb.sql.connect(username=username,
                                        password=password,
                                        hostname=hostname,
                                        database=database)

    def write_df(self, table_name, df, **kwargs):
        mdb = CatapultMdb(self.conn)
        schema_name = kwargs.get("schema", "public")
        dtype = kwargs.get("dtype", None)
        table_schema_only = kwargs.get("table_schema_only", None)
        data_only = kwargs.get("data_only", None)
        return mdb.to_sql(df, schema_name, table_name, dtype, table_schema_only=table_schema_only, data_only=data_only)

    def add_pk(self, *args):
        mdb = CatapultMdb(self.conn)
        return mdb.add_pk(*args)
